"""Module containing functions to convert prediction outputs to desired form."""

from typing import Any, Literal, Union, overload

import numpy as np
from numpy.typing import NDArray

from ..config.tile_information import TileInformation
from .stitch_prediction import stitch_prediction


def convert_outputs(predictions: list[Any], tiled: bool) -> list[NDArray]:
    """
    Convert the Lightning trainer outputs to the desired form.

    This method allows stitching back together tiled predictions.

    Parameters
    ----------
    predictions : list
        Predictions that are output from `Trainer.predict`.
    tiled : bool
        Whether the predictions are tiled.

    Returns
    -------
    list of numpy.ndarray or numpy.ndarray
        list of arrays with the axes SC(Z)YX. If there is only 1 output it will not
        be in a list.
    """
    if len(predictions) == 0:
        return predictions

    # this layout is to stop mypy complaining
    if tiled:
        predictions_comb = combine_batches(predictions, tiled)
        predictions_output = stitch_prediction(*predictions_comb)
    else:
        predictions_output = combine_batches(predictions, tiled)

    return predictions_output


# for mypy
@overload
def combine_batches(  # numpydoc ignore=GL08
    predictions: list[Any], tiled: Literal[True]
) -> tuple[list[NDArray], list[TileInformation]]: ...


# for mypy
@overload
def combine_batches(  # numpydoc ignore=GL08
    predictions: list[Any], tiled: Literal[False]
) -> list[NDArray]: ...


# for mypy
@overload
def combine_batches(  # numpydoc ignore=GL08
    predictions: list[Any], tiled: Union[bool, Literal[True], Literal[False]]
) -> Union[list[NDArray], tuple[list[NDArray], list[TileInformation]]]: ...


def combine_batches(
    predictions: list[Any], tiled: bool
) -> Union[list[NDArray], tuple[list[NDArray], list[TileInformation]]]:
    """
    If predictions are in batches, they will be combined.

    Parameters
    ----------
    predictions : list
        Predictions that are output from `Trainer.predict`.
    tiled : bool
        Whether the predictions are tiled.

    Returns
    -------
    (list of numpy.ndarray) or tuple of (list of numpy.ndarray, list of TileInformation)
        Combined batches.
    """
    if tiled:
        return _combine_tiled_batches(predictions)
    else:
        return _combine_array_batches(predictions)


def _combine_tiled_batches(
    predictions: list[tuple[NDArray, list[TileInformation]]]
) -> tuple[list[NDArray], list[TileInformation]]:
    """
    Combine batches from tiled output.

    Parameters
    ----------
    predictions : list of (numpy.ndarray, list of TileInformation)
        Predictions that are output from `Trainer.predict`. For tiled batches, this is
        a list of tuples. The first element of the tuples is the prediction output of
        tiles with dimension (B, C, (Z), Y, X), where B is batch size. The second
        element of the tuples is a list of TileInformation objects of length B.

    Returns
    -------
    tuple of (list of numpy.ndarray, list of TileInformation)
        Combined batches.
    """
    # turn list of lists into single list
    tile_infos = [
        tile_info for _, tile_info_list in predictions for tile_info in tile_info_list
    ]
    prediction_tiles: list[NDArray] = _combine_array_batches(
        [preds for preds, _ in predictions]
    )
    return prediction_tiles, tile_infos


def _combine_array_batches(predictions: list[NDArray]) -> list[NDArray]:
    """
    Combine batches of arrays.

    Parameters
    ----------
    predictions : list
        Prediction arrays that are output from `Trainer.predict`. A list of arrays that
        have dimensions (B, C, (Z), Y, X), where B is batch size.

    Returns
    -------
    list of numpy.ndarray
        A list of arrays with dimensions (1, C, (Z), Y, X).
    """
    prediction_concat: NDArray = np.concatenate(predictions, axis=0)
    prediction_split = np.split(prediction_concat, prediction_concat.shape[0], axis=0)
    return prediction_split



"""PyTorch lightning utilities."""

from pathlib import Path
from typing import Union


def read_csv_logger(experiment_name: str, log_folder: Union[str, Path]) -> dict:
    """Return the loss curves from the csv logs.

    Parameters
    ----------
    experiment_name : str
        Name of the experiment.
    log_folder : Path or str
        Path to the folder containing the csv logs.

    Returns
    -------
    dict
        Dictionary containing the loss curves, with keys "train_epoch", "val_epoch",
        "train_loss" and "val_loss".
    """
    path = Path(log_folder) / experiment_name

    # find the most recent of version_* folders
    versions = [int(v.name.split("_")[-1]) for v in path.iterdir() if v.is_dir()]
    version = max(versions)

    path_log = path / f"version_{version}" / "metrics.csv"

    epochs = []
    train_losses_tmp = []
    val_losses_tmp = []
    with open(path_log) as f:
        lines = f.readlines()

        for single_line in lines[1:]:
            epoch, _, train_loss, _, val_loss = single_line.strip().split(",")

            epochs.append(epoch)
            train_losses_tmp.append(train_loss)
            val_losses_tmp.append(val_loss)

    # train and val are not logged on the same row and can have different lengths
    train_epoch = [
        int(epochs[i]) for i in range(len(epochs)) if train_losses_tmp[i] != ""
    ]
    val_epoch = [int(epochs[i]) for i in range(len(epochs)) if val_losses_tmp[i] != ""]
    train_losses = [float(loss) for loss in train_losses_tmp if loss != ""]
    val_losses = [float(loss) for loss in val_losses_tmp if loss != ""]

    return {
        "train_epoch": train_epoch,
        "val_epoch": val_epoch,
        "train_loss": train_losses,
        "val_loss": val_losses,
    }



import json
from pathlib import Path
from typing import Union

import numpy as np
import pytest
import torch
from pydantic import BaseModel, ConfigDict

from careamics.config.likelihood_model import Tensor
from careamics.config.nm_model import Array


class MyArray(BaseModel):

    model_config = ConfigDict(arbitrary_types_allowed=True)

    arr: Array


class MyTensor(BaseModel):

    model_config = ConfigDict(arbitrary_types_allowed=True)

    arr: Tensor


@pytest.mark.parametrize("arr", [np.array([1, 2]), torch.tensor([1, 2])])
def test_serialize_array(arr: Union[np.ndarray, torch.Tensor]):
    """Test array_to_json function."""
    arr_model = MyArray(arr=arr)
    assert arr_model.model_dump() == {"arr": "[1, 2]"}


@pytest.mark.parametrize("arr", [np.array([1, 2]), torch.tensor([1, 2])])
def test_serialize_tensor(arr: Union[np.ndarray, torch.Tensor]):
    """Test array_to_json function."""
    arr_model = MyTensor(arr=arr)
    assert arr_model.model_dump() == {"arr": "[1, 2]"}


def test_deserialize_array(tmp_path: Path):
    """Test list_to_numpy function."""
    arr_model = MyArray(arr=np.array([1, 2]))
    # save to JSON
    with open(tmp_path / "array_config.json", "w") as f:
        f.write(arr_model.model_dump_json())
    # load from JSON
    with open(tmp_path / "array_config.json") as f:
        config = json.load(f)
    new_arr_model = MyArray(**config)
    assert np.array_equal(new_arr_model.arr, np.array([1, 2]))


def test_deserialize_tensor(tmp_path: Path):
    """Test list_to_tensor function."""
    arr_model = MyTensor(arr=torch.tensor([1, 2]))
    # save to JSON
    with open(tmp_path / "tensor_config.json", "w") as f:
        f.write(arr_model.model_dump_json())
    # load from JSON
    with open(tmp_path / "tensor_config.json") as f:
        config = json.load(f)
    new_arr_model = MyTensor(**config)
    assert torch.equal(new_arr_model.arr, torch.tensor([1, 2]))



from pytorch_lightning.callbacks import EarlyStopping, ModelCheckpoint

from careamics.config.callback_model import CheckpointModel, EarlyStoppingModel


def test_defaults_checkpoint():
    """Test that a default CheckpointModel can initialize a ModelCheckpoint callback."""
    config = CheckpointModel()
    ModelCheckpoint(**config.model_dump())


def test_defaults_early_stopping():
    """Test that a default EarlyStoppingModel can initialize an EarlyStopping
    callback."""
    config = EarlyStoppingModel()
    EarlyStopping(**config.model_dump())



from pathlib import Path

import numpy as np
import pytest
import tifffile

from careamics.config.support import SupportedData
from careamics.dataset.dataset_utils import (
    get_files_size,
    list_files,
    validate_source_target_files,
)


def test_get_files_size_tiff(tmp_path: Path):
    """Test getting size of multiple TIFF files."""
    # create array
    image = np.ones((10, 10))

    # save array to tiff
    path1 = tmp_path / "test1.tif"
    tifffile.imwrite(path1, image)

    path2 = tmp_path / "test2.tiff"
    tifffile.imwrite(path2, image)

    # save text file
    path3 = tmp_path / "test3.txt"
    path3.write_text("test")

    # save file in subdirectory
    subdirectory = tmp_path / "subdir"
    subdirectory.mkdir()
    path4 = subdirectory / "test3.tif"
    tifffile.imwrite(path4, image)

    # create file list
    files = [path1, path2, path4]

    # get files size
    size = get_files_size(files)
    assert size > 0


def test_list_single_file_tiff(tmp_path: Path):
    """Test listing a single TIFF file."""
    # create array
    image = np.ones((10, 10))

    # save array to tiff
    path = tmp_path / "test.tif"
    tifffile.imwrite(path, image)

    # list file using parent directory
    files = list_files(tmp_path, SupportedData.TIFF)
    assert len(files) == 1
    assert files[0] == path

    # list file using file path
    files = list_files(path, SupportedData.TIFF)
    assert len(files) == 1
    assert files[0] == path


def test_list_multiple_files_tiff(tmp_path: Path):
    """Test listing multiple TIFF files in subdirectories with additional files."""
    # create array
    image = np.ones((10, 10))

    # save array to /npy
    path1 = tmp_path / "test1.tif"
    tifffile.imwrite(path1, image)

    path2 = tmp_path / "test2.tif"
    tifffile.imwrite(path2, image)

    # save text file
    path3 = tmp_path / "test3.txt"
    path3.write_text("test")

    # save file in subdirectory
    subdirectory = tmp_path / "subdir"
    subdirectory.mkdir()
    path4 = subdirectory / "test3.tif"
    tifffile.imwrite(path4, image)

    # create file list
    ref_files = [path1, path2, path4]

    # list files using parent directory
    files = list_files(tmp_path, SupportedData.TIFF)
    assert len(files) == 3
    assert set(files) == set(ref_files)

    # test that the files are sorted
    assert files != ref_files
    assert files == sorted(ref_files)


def test_list_single_file_custom(tmp_path):
    """Test listing a single custom file."""
    # create array
    image = np.ones((10, 10))

    # save as .npy
    path = tmp_path / "custom.npy"
    np.save(path, image)

    # list files using parent directory
    files = list_files(tmp_path, SupportedData.CUSTOM)
    assert len(files) == 1
    assert files[0] == path

    # list files using file path
    files = list_files(path, SupportedData.CUSTOM)
    assert len(files) == 1
    assert files[0] == path


def test_list_multiple_files_custom(tmp_path: Path):
    """Test listing multiple custom files in subdirectories with additional files."""
    # create array
    image = np.ones((10, 10))

    # save array to /npy
    path1 = tmp_path / "test1.npy"
    np.save(path1, image)

    path2 = tmp_path / "test2.npy"
    np.save(path2, image)

    # save text file
    path3 = tmp_path / "test3.txt"
    path3.write_text("test")

    # save file in subdirectory
    subdirectory = tmp_path / "subdir"
    subdirectory.mkdir()
    path4 = subdirectory / "test3.npy"
    np.save(path4, image)

    # create file list (even the text file is selected)
    ref_files = [path1, path2, path3, path4]

    # list files using parent directory
    files = list_files(tmp_path, SupportedData.CUSTOM)
    assert len(files) == 4
    assert set(files) == set(ref_files)

    # list files using the file extension filter
    files = list_files(tmp_path, SupportedData.CUSTOM, "*.npy")
    assert len(files) == 3
    assert set(files) == {path1, path2, path4}


def test_validate_source_target_files(tmp_path: Path):
    """Test that it passes for two folders with same number of files and same names."""
    # create two subfolders
    src = tmp_path / "src"
    src.mkdir()

    tar = tmp_path / "tar"
    tar.mkdir()

    # populate with files
    filename_1 = "test1.txt"
    filename_2 = "test2.txt"

    (tmp_path / "src" / filename_1).write_text("test")
    (tmp_path / "tar" / filename_1).write_text("test")

    (tmp_path / "src" / filename_2).write_text("test")
    (tmp_path / "tar" / filename_2).write_text("test")

    # list files
    src_files = list_files(src, SupportedData.CUSTOM)
    tar_files = list_files(tar, SupportedData.CUSTOM)

    # validate files
    validate_source_target_files(src_files, tar_files)


def test_validate_source_target_files_wrong_names(tmp_path: Path):
    """Test that an error is raised if filenames are different."""
    # create two subfolders
    src = tmp_path / "src"
    src.mkdir()

    tar = tmp_path / "tar"
    tar.mkdir()

    # populate with files
    filename_1 = "test1.txt"
    filename_2 = "test2.txt"
    filename_3 = "test3.txt"

    (tmp_path / "src" / filename_1).write_text("test")
    (tmp_path / "tar" / filename_1).write_text("test")

    (tmp_path / "src" / filename_2).write_text("test")
    (tmp_path / "tar" / filename_3).write_text("test")

    # list files
    src_files = list_files(src, SupportedData.CUSTOM)
    tar_files = list_files(tar, SupportedData.CUSTOM)

    # validate files
    with pytest.raises(ValueError):
        validate_source_target_files(src_files, tar_files)


def test_validate_source_target_files_wrong_number(tmp_path: Path):
    """Test that an error is raised if filenames are different."""
    # create two subfolders
    src = tmp_path / "src"
    src.mkdir()

    tar = tmp_path / "tar"
    tar.mkdir()

    # populate with files
    filename_1 = "test1.txt"
    filename_2 = "test2.txt"

    (tmp_path / "src" / filename_1).write_text("test")
    (tmp_path / "tar" / filename_1).write_text("test")

    (tmp_path / "src" / filename_2).write_text("test")

    # list files
    src_files = list_files(src, SupportedData.CUSTOM)
    tar_files = list_files(tar, SupportedData.CUSTOM)

    # validate files
    with pytest.raises(ValueError):
        validate_source_target_files(src_files, tar_files)



from collections.abc import Sequence
from typing import Callable, Union

import numpy as np
from numpy.typing import NDArray

from .config import DatasetConfig
from .lc_dataset import LCMultiChDloader
from .multich_dataset import MultiChDloader
from .types import DataSplitType


class TwoChannelData(Sequence):
    """
    each element in data_arr should be a N*H*W array
    """

    def __init__(self, data_arr1, data_arr2, paths_data1=None, paths_data2=None):
        assert len(data_arr1) == len(data_arr2)
        self.paths1 = paths_data1
        self.paths2 = paths_data2

        self._data = []
        for i in range(len(data_arr1)):
            assert data_arr1[i].shape == data_arr2[i].shape
            assert (
                len(data_arr1[i].shape) == 3
            ), f"Each element in data arrays should be a N*H*W, but {data_arr1[i].shape}"
            self._data.append(
                np.concatenate(
                    [data_arr1[i][..., None], data_arr2[i][..., None]], axis=-1
                )
            )

    def __len__(self):
        n = 0
        for x in self._data:
            n += x.shape[0]
        return n

    def __getitem__(self, idx):
        n = 0
        for dataidx, x in enumerate(self._data):
            if idx < n + x.shape[0]:
                if self.paths1 is None:
                    return x[idx - n], None
                else:
                    return x[idx - n], (self.paths1[dataidx], self.paths2[dataidx])
            n += x.shape[0]
        raise IndexError("Index out of range")


class MultiChannelData(Sequence):
    """
    each element in data_arr should be a N*H*W array
    """

    def __init__(self, data_arr, paths=None):
        self.paths = paths

        self._data = data_arr

    def __len__(self):
        n = 0
        for x in self._data:
            n += x.shape[0]
        return n

    def __getitem__(self, idx):
        n = 0
        for dataidx, x in enumerate(self._data):
            if idx < n + x.shape[0]:
                if self.paths is None:
                    return x[idx - n], None
                else:
                    return x[idx - n], (self.paths[dataidx])
            n += x.shape[0]
        raise IndexError("Index out of range")


class SingleFileLCDset(LCMultiChDloader):
    def __init__(
        self,
        preloaded_data: NDArray,
        data_config: DatasetConfig,
        fpath: str,
        load_data_fn: Callable,
        val_fraction=None,
        test_fraction=None,
    ):
        self._preloaded_data = preloaded_data
        super().__init__(
            data_config,
            fpath,
            load_data_fn=load_data_fn,
            val_fraction=val_fraction,
            test_fraction=test_fraction,
        )

    @property
    def data_path(self):
        return self._fpath

    def rm_bkground_set_max_val_and_upperclip_data(self, max_val, datasplit_type):
        pass

    def load_data(
        self,
        data_config: DatasetConfig,
        datasplit_type: DataSplitType,
        load_data_fn: Callable,
        val_fraction=None,
        test_fraction=None,
        allow_generation=None,
    ):
        self._data = self._preloaded_data
        assert "channel_1" not in data_config or isinstance(data_config.channel_1, str)
        assert "channel_2" not in data_config or isinstance(data_config.channel_2, str)
        assert "channel_3" not in data_config or isinstance(data_config.channel_3, str)
        self._loaded_data_preprocessing(data_config)


class SingleFileDset(MultiChDloader):
    def __init__(
        self,
        preloaded_data: NDArray,
        data_config: DatasetConfig,
        fpath: str,
        load_data_fn: Callable,
        val_fraction=None,
        test_fraction=None,
    ):
        self._preloaded_data = preloaded_data
        super().__init__(
            data_config,
            fpath,
            load_data_fn=load_data_fn,
            val_fraction=val_fraction,
            test_fraction=test_fraction,
        )

    def rm_bkground_set_max_val_and_upperclip_data(self, max_val, datasplit_type):
        pass

    @property
    def data_path(self):
        return self._fpath

    def load_data(
        self,
        data_config: DatasetConfig,
        datasplit_type: DataSplitType,
        load_data_fn: Callable[..., NDArray],
        val_fraction=None,
        test_fraction=None,
        allow_generation=None,
    ):
        self._data = self._preloaded_data
        assert (
            "channel_1" not in data_config
        ), "Outdated config file. Please remove channel_1, channel_2, channel_3 from the config file."
        assert (
            "channel_2" not in data_config
        ), "Outdated config file. Please remove channel_1, channel_2, channel_3 from the config file."
        assert (
            "channel_3" not in data_config
        ), "Outdated config file. Please remove channel_1, channel_2, channel_3 from the config file."
        self._loaded_data_preprocessing(data_config)


class MultiFileDset:
    """
    Here, we handle dataset having multiple files. Each file can have a different spatial dimension and number of frames (Z stack).
    """

    def __init__(
        self,
        data_config: DatasetConfig,
        fpath: str,
        load_data_fn: Callable[..., Union[TwoChannelData, MultiChannelData]],
        val_fraction=None,
        test_fraction=None,
    ):
        self._fpath = fpath
        data: Union[TwoChannelData, MultiChannelData] = load_data_fn(
            data_config,
            self._fpath,
            data_config.datasplit_type,
            val_fraction=val_fraction,
            test_fraction=test_fraction,
        )
        self.dsets = []

        for i in range(len(data)):
            prefetched_data, fpath_tuple = data[i]
            if (
                data_config.multiscale_lowres_count is not None
                and data_config.multiscale_lowres_count > 1
            ):

                self.dsets.append(
                    SingleFileLCDset(
                        prefetched_data[None],
                        data_config,
                        fpath_tuple,
                        load_data_fn,
                        val_fraction=val_fraction,
                        test_fraction=test_fraction,
                    )
                )

            else:
                self.dsets.append(
                    SingleFileDset(
                        prefetched_data[None],
                        data_config,
                        fpath_tuple,
                        load_data_fn,
                        val_fraction=val_fraction,
                        test_fraction=test_fraction,
                    )
                )

        self.rm_bkground_set_max_val_and_upperclip_data(
            data_config.max_val, data_config.datasplit_type
        )
        count = 0
        avg_height = 0
        avg_width = 0
        for dset in self.dsets:
            shape = dset.get_data_shape()
            avg_height += shape[1]
            avg_width += shape[2]
            count += shape[0]

        avg_height = int(avg_height / len(self.dsets))
        avg_width = int(avg_width / len(self.dsets))
        print(
            f"{self.__class__.__name__} avg height: {avg_height}, avg width: {avg_width}, count: {count}"
        )

    def rm_bkground_set_max_val_and_upperclip_data(self, max_val, datasplit_type):
        self.set_max_val(max_val, datasplit_type)
        self.upperclip_data()

    def set_mean_std(self, mean_val, std_val):
        for dset in self.dsets:
            dset.set_mean_std(mean_val, std_val)

    def get_mean_std(self):
        return self.dsets[0].get_mean_std()

    def compute_max_val(self):
        max_val_arr = []
        for dset in self.dsets:
            max_val_arr.append(dset.compute_max_val())
        return np.max(max_val_arr)

    def set_max_val(self, max_val, datasplit_type):
        if datasplit_type == DataSplitType.Train:
            assert max_val is None
            max_val = self.compute_max_val()
        for dset in self.dsets:
            dset.set_max_val(max_val, datasplit_type)

    def upperclip_data(self):
        for dset in self.dsets:
            dset.upperclip_data()

    def get_max_val(self):
        return self.dsets[0].get_max_val()

    def get_img_sz(self):
        return self.dsets[0].get_img_sz()

    def set_img_sz(self, image_size, grid_size):
        for dset in self.dsets:
            dset.set_img_sz(image_size, grid_size)

    def compute_mean_std(self):
        cur_mean = {"target": 0, "input": 0}
        cur_std = {"target": 0, "input": 0}
        for dset in self.dsets:
            mean, std = dset.compute_mean_std()
            cur_mean["target"] += mean["target"]
            cur_mean["input"] += mean["input"]

            cur_std["target"] += std["target"]
            cur_std["input"] += std["input"]

        cur_mean["target"] /= len(self.dsets)
        cur_mean["input"] /= len(self.dsets)
        cur_std["target"] /= len(self.dsets)
        cur_std["input"] /= len(self.dsets)
        return cur_mean, cur_std

    def compute_individual_mean_std(self):
        cum_mean = 0
        cum_std = 0
        for dset in self.dsets:
            mean, std = dset.compute_individual_mean_std()
            cum_mean += mean
            cum_std += std
        return cum_mean / len(self.dsets), cum_std / len(self.dsets)

    def get_num_frames(self):
        return len(self.dsets)

    def reduce_data(
        self, t_list=None, h_start=None, h_end=None, w_start=None, w_end=None
    ):
        assert h_start is None
        assert h_end is None
        assert w_start is None
        assert w_end is None
        self.dsets = [self.dsets[t] for t in t_list]
        print(
            f"[{self.__class__.__name__}] Data reduced. New data count: {len(self.dsets)}"
        )

    def __len__(self):
        out = 0
        for dset in self.dsets:
            out += len(dset)
        return out

    def __getitem__(self, idx):
        cum_len = 0
        for dset in self.dsets:
            cum_len += len(dset)
            if idx < cum_len:
                rel_idx = idx - (cum_len - len(dset))
                return dset[rel_idx]

        raise IndexError("Index out of range")



import pytest

from careamics.config.inference_model import InferenceConfig


@pytest.mark.parametrize("ext", ["nd2", "jpg", "png ", "zarr", "npy"])
def test_wrong_extensions(minimum_inference: dict, ext: str):
    """Test that supported model raises ValueError for unsupported extensions."""
    minimum_inference["data_type"] = ext

    # instantiate InferenceModel model
    with pytest.raises(ValueError):
        InferenceConfig(**minimum_inference)


def test_mean_std_both_specified_or_none(minimum_inference: dict):
    """Test error raising when setting mean and std."""
    # Errors if both are None
    minimum_inference["image_means"] = []
    minimum_inference["image_stds"] = []
    with pytest.raises(ValueError):
        InferenceConfig(**minimum_inference)

    # Error if only mean is defined
    minimum_inference["image_means"] = [10.4]
    with pytest.raises(ValueError):
        InferenceConfig(**minimum_inference)

    # Error if only std is defined
    minimum_inference.pop("image_means")
    minimum_inference["image_stds"] = [10.4]
    with pytest.raises(ValueError):
        InferenceConfig(**minimum_inference)

    # No error if both are specified
    minimum_inference["image_means"] = [10.4]
    minimum_inference["image_stds"] = [10.4]
    InferenceConfig(**minimum_inference)


def test_tile_size(minimum_inference: dict):
    """Test that non-zero even patch size are accepted."""
    # no tiling
    prediction_model = InferenceConfig(**minimum_inference)

    # 2D
    minimum_inference["tile_size"] = [16, 8]
    minimum_inference["tile_overlap"] = [2, 2]
    minimum_inference["axes"] = "YX"

    prediction_model = InferenceConfig(**minimum_inference)
    assert prediction_model.tile_size == minimum_inference["tile_size"]
    assert prediction_model.tile_overlap == minimum_inference["tile_overlap"]

    # 3D
    minimum_inference["tile_size"] = [16, 8, 32]
    minimum_inference["tile_overlap"] = [2, 2, 2]
    minimum_inference["axes"] = "ZYX"

    prediction_model = InferenceConfig(**minimum_inference)
    assert prediction_model.tile_size == minimum_inference["tile_size"]
    assert prediction_model.tile_overlap == minimum_inference["tile_overlap"]


@pytest.mark.parametrize(
    "tile_size", [[12], [0, 12, 12], [12, 12, 13], [12, 12, 12, 12]]
)
def test_wrong_tile_size(minimum_inference: dict, tile_size):
    """Test that wrong patch sizes are not accepted (zero or odd, dims 1 or > 3)."""
    minimum_inference["axes"] = "ZYX" if len(tile_size) == 3 else "YX"
    minimum_inference["tile_size"] = tile_size

    with pytest.raises(ValueError):
        InferenceConfig(**minimum_inference)


@pytest.mark.parametrize(
    "tile_size, tile_overlap", [([12, 12], [2, 2, 2]), ([12, 12, 12], [14, 2, 2])]
)
def test_wrong_tile_overlap(minimum_inference: dict, tile_size, tile_overlap):
    """Test that wrong patch sizes are not accepted (zero or odd, dims 1 or > 3)."""
    minimum_inference["axes"] = "ZYX" if len(tile_size) == 3 else "YX"
    minimum_inference["tile_size"] = tile_size
    minimum_inference["tile_overlap"] = tile_overlap

    with pytest.raises(ValueError):
        InferenceConfig(**minimum_inference)


def test_set_3d(minimum_inference: dict):
    """Test that 3D can be set."""
    minimum_inference["tile_size"] = [64, 64]
    minimum_inference["tile_overlap"] = [32, 32]

    pred = InferenceConfig(**minimum_inference)
    assert "Z" not in pred.axes
    assert len(pred.tile_size) == 2
    assert len(pred.tile_overlap) == 2

    # error if changing Z manually
    with pytest.raises(ValueError):
        pred.axes = "ZYX"

    # or patch size
    pred = InferenceConfig(**minimum_inference)
    with pytest.raises(ValueError):
        pred.tile_size = [64, 64, 64]

    with pytest.raises(ValueError):
        pred.tile_overlap = [64, 64, 64]

    # set 3D
    pred = InferenceConfig(**minimum_inference)
    pred.set_3D("ZYX", [64, 64, 64], [32, 32, 32])
    assert "Z" in pred.axes
    assert len(pred.tile_size) == 3
    assert len(pred.tile_overlap) == 3



""""N2V Algorithm configuration."""

from typing import Annotated, Literal

from bioimageio.spec.generic.v0_3 import CiteEntry
from pydantic import AfterValidator, ConfigDict, model_validator
from typing_extensions import Self

from careamics.config.architectures import UNetModel
from careamics.config.support import SupportedPixelManipulation, SupportedStructAxis
from careamics.config.transformations import N2VManipulateModel
from careamics.config.validators import (
    model_matching_in_out_channels,
    model_without_final_activation,
)

from .unet_algorithm_model import UNetBasedAlgorithm

N2V = "Noise2Void"
N2V2 = "N2V2"
STRUCT_N2V = "StructN2V"
STRUCT_N2V2 = "StructN2V2"

N2V_REF = CiteEntry(
    text='Krull, A., Buchholz, T.O. and Jug, F., 2019. "Noise2Void - Learning '
    'denoising from single noisy images". In Proceedings of the IEEE/CVF '
    "conference on computer vision and pattern recognition (pp. 2129-2137).",
    doi="10.1109/cvpr.2019.00223",
)

N2V2_REF = CiteEntry(
    text="Höck, E., Buchholz, T.O., Brachmann, A., Jug, F. and Freytag, A., "
    '2022. "N2V2 - Fixing Noise2Void checkerboard artifacts with modified '
    'sampling strategies and a tweaked network architecture". In European '
    "Conference on Computer Vision (pp. 503-518).",
    doi="10.1007/978-3-031-25069-9_33",
)

STRUCTN2V_REF = CiteEntry(
    text="Broaddus, C., Krull, A., Weigert, M., Schmidt, U. and Myers, G., 2020."
    '"Removing structured noise with self-supervised blind-spot '
    'networks". In 2020 IEEE 17th International Symposium on Biomedical '
    "Imaging (ISBI) (pp. 159-163).",
    doi="10.1109/isbi45749.2020.9098336",
)

N2V_DESCRIPTION = (
    "Noise2Void is a UNet-based self-supervised algorithm that "
    "uses blind-spot training to denoise images. In short, in every "
    "patches during training, random pixels are selected and their "
    "value replaced by a neighboring pixel value. The network is then "
    "trained to predict the original pixel value. The algorithm "
    "relies on the continuity of the signal (neighboring pixels have "
    "similar values) and the pixel-wise independence of the noise "
    "(the noise in a pixel is not correlated with the noise in "
    "neighboring pixels)."
)

N2V2_DESCRIPTION = (
    "N2V2 is a variant of Noise2Void. "
    + N2V_DESCRIPTION
    + "\nN2V2 introduces blur-pool layers and removed skip "
    "connections in the UNet architecture to remove checkboard "
    "artefacts, a common artefacts ocurring in Noise2Void."
)

STR_N2V_DESCRIPTION = (
    "StructN2V is a variant of Noise2Void. "
    + N2V_DESCRIPTION
    + "\nStructN2V uses a linear mask (horizontal or vertical) to replace "
    "the pixel values of neighbors of the masked pixels by a random "
    "value. Such masking allows removing 1D structured noise from the "
    "the images, the main failure case of the original N2V."
)

STR_N2V2_DESCRIPTION = (
    "StructN2V2 is a a variant of Noise2Void that uses both "
    "structN2V and N2V2. "
    + N2V_DESCRIPTION
    + "\nStructN2V2 uses a linear mask (horizontal or vertical) to replace "
    "the pixel values of neighbors of the masked pixels by a random "
    "value. Such masking allows removing 1D structured noise from the "
    "the images, the main failure case of the original N2V."
    "\nN2V2 introduces blur-pool layers and removed skip connections in "
    "the UNet architecture to remove checkboard artefacts, a common "
    "artefacts ocurring in Noise2Void."
)


class N2VAlgorithm(UNetBasedAlgorithm):
    """N2V Algorithm configuration."""

    model_config = ConfigDict(validate_assignment=True)

    algorithm: Literal["n2v"] = "n2v"
    """N2V Algorithm name."""

    loss: Literal["n2v"] = "n2v"
    """N2V loss function."""

    n2v_config: N2VManipulateModel = N2VManipulateModel()

    model: Annotated[
        UNetModel,
        AfterValidator(model_matching_in_out_channels),
        AfterValidator(model_without_final_activation),
    ]

    @model_validator(mode="after")
    def validate_n2v2(self) -> Self:
        """Validate that the N2V2 strategy and models are set correctly.

        Returns
        -------
        Self
            The validateed configuration.

        Raises
        ------
        ValueError
            If N2V2 is used with the wrong pixel manipulation strategy.
        """
        if self.model.n2v2:
            if self.n2v_config.strategy != SupportedPixelManipulation.MEDIAN.value:
                raise ValueError(
                    f"N2V2 can only be used with the "
                    f"{SupportedPixelManipulation.MEDIAN} pixel manipulation strategy. "
                    f"Change the `strategy` parameters in `n2v_config` to "
                    f"{SupportedPixelManipulation.MEDIAN}."
                )
        else:
            if self.n2v_config.strategy != SupportedPixelManipulation.UNIFORM.value:
                raise ValueError(
                    f"N2V can only be used with the "
                    f"{SupportedPixelManipulation.UNIFORM} pixel manipulation strategy."
                    f" Change the `strategy` parameters in `n2v_config` to "
                    f"{SupportedPixelManipulation.UNIFORM}."
                )
        return self

    def set_n2v2(self, use_n2v2: bool) -> None:
        """
        Set the configuration to use N2V2 or the vanilla Noise2Void.

        This method ensures that N2V2 is set correctly and remain coherent, as opposed
        to setting the different parameters individually.

        Parameters
        ----------
        use_n2v2 : bool
            Whether to use N2V2.
        """
        if use_n2v2:
            self.n2v_config.strategy = SupportedPixelManipulation.MEDIAN.value
            self.model.n2v2 = True
        else:
            self.n2v_config.strategy = SupportedPixelManipulation.UNIFORM.value
            self.model.n2v2 = False

    def is_struct_n2v(self) -> bool:
        """Check if the configuration is using structN2V.

        Returns
        -------
        bool
            Whether the configuration is using structN2V.
        """
        return self.n2v_config.struct_mask_axis != SupportedStructAxis.NONE.value

    def get_algorithm_friendly_name(self) -> str:
        """
        Get the friendly name of the algorithm.

        Returns
        -------
        str
            Friendly name.
        """
        use_n2v2 = self.model.n2v2
        use_structN2V = self.is_struct_n2v()

        if use_n2v2 and use_structN2V:
            return STRUCT_N2V2
        elif use_n2v2:
            return N2V2
        elif use_structN2V:
            return STRUCT_N2V
        else:
            return N2V

    def get_algorithm_keywords(self) -> list[str]:
        """
        Get algorithm keywords.

        Returns
        -------
        list[str]
            List of keywords.
        """
        use_n2v2 = self.model.n2v2
        use_structN2V = self.is_struct_n2v()

        keywords = [
            "denoising",
            "restoration",
            "UNet",
            "3D" if self.model.is_3D() else "2D",
            "CAREamics",
            "pytorch",
            N2V,
        ]

        if use_n2v2:
            keywords.append(N2V2)
        if use_structN2V:
            keywords.append(STRUCT_N2V)

        return keywords

    def get_algorithm_references(self) -> str:
        """
        Get the algorithm references.

        This is used to generate the README of the BioImage Model Zoo export.

        Returns
        -------
        str
            Algorithm references.
        """
        use_n2v2 = self.model.n2v2
        use_structN2V = self.is_struct_n2v()

        references = [
            N2V_REF.text + " doi: " + N2V_REF.doi,
            N2V2_REF.text + " doi: " + N2V2_REF.doi,
            STRUCTN2V_REF.text + " doi: " + STRUCTN2V_REF.doi,
        ]

        # return the (struct)N2V(2) references
        if use_n2v2 and use_structN2V:
            return "\n".join(references)
        elif use_n2v2:
            references.pop(-1)
            return "\n".join(references)
        elif use_structN2V:
            references.pop(-2)
            return "\n".join(references)
        else:
            return references[0]

    def get_algorithm_citations(self) -> list[CiteEntry]:
        """
        Return a list of citation entries of the current algorithm.

        This is used to generate the model description for the BioImage Model Zoo.

        Returns
        -------
        List[CiteEntry]
            List of citation entries.
        """
        use_n2v2 = self.model.n2v2
        use_structN2V = self.is_struct_n2v()

        references = [N2V_REF]

        if use_n2v2:
            references.append(N2V2_REF)

        if use_structN2V:
            references.append(STRUCTN2V_REF)

        return references

    def get_algorithm_description(self) -> str:
        """
        Return a description of the algorithm.

        This method is used to generate the README of the BioImage Model Zoo export.

        Returns
        -------
        str
            Description of the algorithm.
        """
        use_n2v2 = self.model.n2v2
        use_structN2V = self.is_struct_n2v()

        if use_n2v2 and use_structN2V:
            return STR_N2V2_DESCRIPTION
        elif use_n2v2:
            return N2V2_DESCRIPTION
        elif use_structN2V:
            return STR_N2V_DESCRIPTION
        else:
            return N2V_DESCRIPTION



import pytest

from careamics.config.architectures import UNetModel
from careamics.config.validators import (
    model_matching_in_out_channels,
    model_without_final_activation,
    model_without_n2v2,
)


def test_model_without_n2v2():
    """Test the validation of the model without the `n2v2` attribute."""
    model = UNetModel(architecture="UNet", final_activation="None", n2v2=False)
    assert model_without_n2v2(model) == model

    model = UNetModel(architecture="UNet", final_activation="None", n2v2=True)
    with pytest.raises(ValueError):
        model_without_n2v2(model)


def test_model_without_final_activation():
    """Test the validation of the model without the `final_activation` attribute."""
    model = UNetModel(
        architecture="UNet",
        final_activation="None",
    )
    assert model_without_final_activation(model) == model

    model = UNetModel(
        architecture="UNet",
        final_activation="ReLU",
    )
    with pytest.raises(ValueError):
        model_without_final_activation(model)


def test_model_matching_in_out_channels():
    """Test the validation of the model with matching in and out channels."""
    model = UNetModel(
        architecture="UNet",
        in_channels=1,
        num_classes=1,
    )
    assert model_matching_in_out_channels(model) == model

    model = UNetModel(
        architecture="UNet",
        in_channels=1,
        num_classes=2,
    )
    with pytest.raises(ValueError):
        model_matching_in_out_channels(model)



import pytest

from careamics.config import (
    CAREAlgorithm,
    Configuration,
    N2NAlgorithm,
    N2VAlgorithm,
    algorithm_factory,
    create_care_configuration,
    create_n2n_configuration,
    create_n2v_configuration,
)
from careamics.config.configuration_factories import (
    _create_data_configuration,
    _create_supervised_config_dict,
    _create_unet_configuration,
    _list_spatial_augmentations,
)
from careamics.config.support import (
    SupportedAlgorithm,
    SupportedPixelManipulation,
    SupportedStructAxis,
    SupportedTransform,
)
from careamics.config.transformations import (
    N2VManipulateModel,
    XYFlipModel,
    XYRandomRotate90Model,
)


def test_careamics_config_n2v(minimum_n2v_configuration):
    """Test that the N2V configuration is created correctly."""
    config = Configuration(**minimum_n2v_configuration)
    assert config.algorithm_config.algorithm == SupportedAlgorithm.N2V.value


@pytest.mark.parametrize(
    "algorithm", [SupportedAlgorithm.N2N.value, SupportedAlgorithm.CARE.value]
)
def test_careamics_config_supervised(minimum_supervised_configuration, algorithm):
    """Test that the supervised configuration is created correctly."""
    min_config = minimum_supervised_configuration
    min_config["algorithm_config"]["algorithm"] = algorithm

    config = Configuration(**min_config)

    assert config.algorithm_config.algorithm == algorithm


def test_algorithm_factory_n2v(minimum_algorithm_n2v):
    """Test that the N2V configuration is created correctly."""
    algorithm = algorithm_factory(minimum_algorithm_n2v)
    assert isinstance(algorithm, N2VAlgorithm)


@pytest.mark.parametrize("algorithm", ["n2n", "care"])
def test_algorithm_factory_supervised(minimum_algorithm_supervised, algorithm):
    """Test that the supervised configuration is created correctly."""
    min_config = minimum_algorithm_supervised
    min_config["algorithm"] = algorithm

    algorithm_config = algorithm_factory(min_config)

    exp_class = N2NAlgorithm if algorithm == "n2n" else CAREAlgorithm
    assert isinstance(algorithm_config, exp_class)


def test_list_aug_default():
    """Test that the default augmentations are present."""
    list_aug = _list_spatial_augmentations(augmentations=None)

    assert len(list_aug) == 2
    assert list_aug[0].name == SupportedTransform.XY_FLIP.value
    assert list_aug[1].name == SupportedTransform.XY_RANDOM_ROTATE90.value


def test_list_aug_no_aug():
    """Test that disabling augmentation results in empty transform list."""
    list_aug = _list_spatial_augmentations(augmentations=[])
    assert len(list_aug) == 0


def test_list_aug_error_duplicate_transforms():
    """Test that an error is raised when there are duplicate transforms."""
    with pytest.raises(ValueError):
        _list_spatial_augmentations(
            augmentations=[XYFlipModel(), XYRandomRotate90Model(), XYFlipModel()],
        )


def test_list_aug_error_wrong_transform():
    """Test that an error is raised when the wrong transform is passed."""
    with pytest.raises(ValueError):
        _list_spatial_augmentations(
            augmentations=[XYFlipModel(), N2VManipulateModel()],
        )


def test_create_data_configuration_train_dataloader_params(minimum_data):
    """Test that shuffle is added silently to the train_dataloader_params."""
    config_dict = minimum_data
    config_dict["train_dataloader_params"] = {"num_workers": 4}

    config = _create_data_configuration(batch_size=1, augmentations=[], **config_dict)
    assert "shuffle" in config.train_dataloader_params
    assert config.train_dataloader_params["shuffle"]


def test_supervised_configuration_passing_transforms():
    """Test that transforms can be passed to the configuration."""
    config_dict = _create_supervised_config_dict(
        algorithm="n2n",
        experiment_name="test",
        data_type="tiff",
        axes="YX",
        patch_size=[64, 64],
        batch_size=8,
        num_epochs=100,
        augmentations=[XYFlipModel()],
    )
    config = Configuration(**config_dict)

    assert len(config.data_config.transforms) == 1
    assert config.data_config.transforms[0].name == SupportedTransform.XY_FLIP.value


def test_model_creation():
    """Test that the correct parameters are passed to the model."""
    model_kwargs = {
        "depth": 4,
        "conv_dims": 2,
        "n2v2": False,
        "in_channels": 2,
        "num_classes": 5,
        "independent_channels": False,
    }

    # choose different parameters
    axes = "XYZ"
    conv_dims = 3
    in_channels = 3
    num_classes = 4
    independent_channels = True
    use_n2v2 = True

    model = _create_unet_configuration(
        axes=axes,
        n_channels_in=in_channels,
        n_channels_out=num_classes,
        independent_channels=independent_channels,
        use_n2v2=use_n2v2,
        model_params=model_kwargs,
    )

    assert model.depth == model_kwargs["depth"]
    assert model.conv_dims == conv_dims
    assert model.n2v2 == use_n2v2
    assert model.in_channels == in_channels
    assert model.num_classes == num_classes
    assert model.independent_channels == independent_channels


def test_create_configuration():
    """Test that the methods correctly passes all parameters."""
    algorithm = "care"
    experiment_name = "test"
    data_type = "tiff"
    axes = "CYX"
    patch_size = [128, 128]
    batch_size = 8
    num_epochs = 100
    transform_list = [XYFlipModel(), XYRandomRotate90Model()]
    independent_channels = False
    loss = "mse"
    n_channels_in = 2
    n_channels_out = 3
    logger = "tensorboard"
    model_params = {
        "depth": 5,
    }
    train_dataloader_params = {
        "num_workers": 4,
        "shuffle": True,
    }
    val_dataloader_params = {
        "num_workers": 4,
    }

    # instantiate config
    config = Configuration(
        **_create_supervised_config_dict(
            algorithm=algorithm,
            experiment_name=experiment_name,
            data_type=data_type,
            axes=axes,
            patch_size=patch_size,
            batch_size=batch_size,
            num_epochs=num_epochs,
            augmentations=transform_list,
            independent_channels=independent_channels,
            loss=loss,
            n_channels_in=n_channels_in,
            n_channels_out=n_channels_out,
            logger=logger,
            model_params=model_params,
            train_dataloader_params=train_dataloader_params,
            val_dataloader_params=val_dataloader_params,
        )
    )

    assert config.algorithm_config.algorithm == algorithm
    assert config.experiment_name == experiment_name
    assert config.data_config.data_type == data_type
    assert config.data_config.axes == axes
    assert config.data_config.patch_size == patch_size
    assert config.data_config.batch_size == batch_size
    assert config.training_config.num_epochs == num_epochs
    assert config.data_config.transforms == transform_list
    assert config.algorithm_config.model.independent_channels == independent_channels
    assert config.algorithm_config.loss == loss
    assert config.algorithm_config.model.in_channels == n_channels_in
    assert config.algorithm_config.model.num_classes == n_channels_out
    assert config.training_config.logger == logger
    assert config.algorithm_config.model.depth == model_params["depth"]
    assert config.data_config.train_dataloader_params == train_dataloader_params
    assert config.data_config.val_dataloader_params == val_dataloader_params


def test_supervised_configuration_error_with_channel_axes():
    """Test that an error is raised if channels are in axes, but the input channel
    number is not specified."""
    with pytest.raises(ValueError):
        _create_supervised_config_dict(
            algorithm="n2n",
            experiment_name="test",
            data_type="tiff",
            axes="CYX",
            patch_size=[64, 64],
            batch_size=8,
            num_epochs=100,
        )


def test_supervised_configuration_singleton_channel():
    """Test that no error is raised if channels are in axes, and the input channel is
    1."""
    _create_supervised_config_dict(
        algorithm="n2n",
        experiment_name="test",
        data_type="tiff",
        axes="CYX",
        patch_size=[64, 64],
        batch_size=8,
        num_epochs=100,
        n_channels_in=1,
    )


def test_supervised_configuration_no_channel():
    """Test that no error is raised without channel and number of inputs."""
    _create_supervised_config_dict(
        algorithm="n2n",
        experiment_name="test",
        data_type="tiff",
        axes="YX",
        patch_size=[64, 64],
        batch_size=8,
        num_epochs=100,
    )


def test_supervised_configuration_error_without_channel_axes():
    """Test that an error is raised if channels are not in axes, but the input channel
    number is specified and greater than 1."""
    with pytest.raises(ValueError):
        _create_supervised_config_dict(
            algorithm="n2n",
            experiment_name="test",
            data_type="tiff",
            axes="YX",
            patch_size=[64, 64],
            batch_size=8,
            num_epochs=100,
            n_channels_in=2,
        )


def test_supervised_configuration_channels():
    """Test that no error is raised if channels are in axes and the input channel
    are specified."""
    _create_supervised_config_dict(
        algorithm="n2n",
        experiment_name="test",
        data_type="tiff",
        axes="CYX",
        patch_size=[64, 64],
        batch_size=8,
        num_epochs=100,
        n_channels_in=4,
    )


def test_n2n_configuration():
    """Test that N2N configuration can be created."""
    config = create_n2n_configuration(
        experiment_name="test",
        data_type="tiff",
        axes="YX",
        patch_size=[64, 64],
        batch_size=8,
        num_epochs=100,
        train_dataloader_params={"num_workers": 2},
    )
    assert config.algorithm_config.algorithm == "n2n"


def test_n2n_configuration_n_channels():
    """Test the behaviour of the number of channels in and out."""
    n_channels_in = 4
    n_channels_out = 5

    # n_channels_out not specified
    config = create_n2n_configuration(
        experiment_name="test",
        data_type="tiff",
        axes="CYX",
        patch_size=[64, 64],
        batch_size=8,
        num_epochs=100,
        n_channels_in=n_channels_in,
    )
    assert config.algorithm_config.model.in_channels == n_channels_in
    assert config.algorithm_config.model.num_classes == n_channels_in

    # specify n_channels_out
    config = create_n2n_configuration(
        experiment_name="test",
        data_type="tiff",
        axes="CYX",
        patch_size=[64, 64],
        batch_size=8,
        num_epochs=100,
        n_channels_in=n_channels_in,
        n_channels_out=n_channels_out,
    )
    assert config.algorithm_config.model.in_channels == n_channels_in
    assert config.algorithm_config.model.num_classes == n_channels_out


def test_care_configuration():
    """Test that CARE configuration can be created."""
    config = create_care_configuration(
        experiment_name="test",
        data_type="tiff",
        axes="YX",
        patch_size=[64, 64],
        batch_size=8,
        num_epochs=100,
        train_dataloader_params={"num_workers": 2},
    )
    assert config.algorithm_config.algorithm == "care"


def test_care_configuration_n_channels():
    """Test the behaviour of the number of channels in and out."""
    n_channels_in = 4
    n_channels_out = 5

    # n_channels_out not specified
    config = create_care_configuration(
        experiment_name="test",
        data_type="tiff",
        axes="CYX",
        patch_size=[64, 64],
        batch_size=8,
        num_epochs=100,
        n_channels_in=n_channels_in,
    )
    assert config.algorithm_config.model.in_channels == n_channels_in
    assert config.algorithm_config.model.num_classes == n_channels_in

    # specify n_channels_out
    config = create_care_configuration(
        experiment_name="test",
        data_type="tiff",
        axes="CYX",
        patch_size=[64, 64],
        batch_size=8,
        num_epochs=100,
        n_channels_in=n_channels_in,
        n_channels_out=n_channels_out,
    )
    assert config.algorithm_config.model.in_channels == n_channels_in
    assert config.algorithm_config.model.num_classes == n_channels_out


def test_n2v_configuration():
    """Test that N2V configuration can be created."""
    config = create_n2v_configuration(
        experiment_name="test",
        data_type="tiff",
        axes="YX",
        patch_size=[64, 64],
        batch_size=8,
        num_epochs=100,
        train_dataloader_params={"num_workers": 2},
    )
    assert isinstance(config.algorithm_config, N2VAlgorithm)


def test_n2v_configuration_no_aug():
    """Test the default n2v transforms."""
    config = create_n2v_configuration(
        experiment_name="test",
        data_type="tiff",
        axes="YX",
        patch_size=[64, 64],
        batch_size=8,
        num_epochs=100,
        augmentations=[],
    )
    assert config.data_config.transforms == []


def test_n2v_configuration_n2v2_structn2v():
    """Test that N2V manipulate is ignored when explicitely passed."""
    use_n2v2 = True
    roi_size = 15
    masked_pixel_percentage = 0.5
    struct_mask_axis = SupportedStructAxis.HORIZONTAL.value
    struct_n2v_span = 15

    config = create_n2v_configuration(
        experiment_name="test",
        data_type="tiff",
        axes="YX",
        patch_size=[64, 64],
        batch_size=8,
        num_epochs=100,
        use_n2v2=use_n2v2,  # median strategy
        roi_size=roi_size,
        masked_pixel_percentage=masked_pixel_percentage,
        struct_n2v_axis=struct_mask_axis,
        struct_n2v_span=struct_n2v_span,
    )
    assert (
        config.algorithm_config.n2v_config.strategy
        == SupportedPixelManipulation.MEDIAN.value
    )
    assert config.algorithm_config.n2v_config.roi_size == roi_size
    assert (
        config.algorithm_config.n2v_config.masked_pixel_percentage
        == masked_pixel_percentage
    )
    assert config.algorithm_config.n2v_config.struct_mask_axis == struct_mask_axis
    assert config.algorithm_config.n2v_config.struct_mask_span == struct_n2v_span



from collections.abc import Sequence

import numpy as np
import pytest

from careamics.dataset_ng.patching_strategies import (
    SequentialPatchingStrategy,
)


@pytest.mark.parametrize(
    "data_shapes,patch_size,overlap",
    [
        [[(2, 1, 32, 32), (1, 1, 19, 37), (3, 1, 14, 9)], (8, 8), (2, 2)],
        [[(2, 1, 32, 32), (1, 1, 19, 37), (3, 1, 14, 9)], (8, 5), (3, 2)],
        [[(2, 1, 32, 32), (1, 1, 19, 37), (3, 1, 14, 9)], (8, 5), (0, 0)],
        [
            [(2, 1, 32, 32, 32), (1, 1, 19, 37, 23), (3, 1, 14, 9, 12)],
            (8, 8, 8),
            (2, 2, 2),
        ],
        [
            [(2, 1, 32, 32, 32), (1, 1, 19, 37, 23), (3, 1, 14, 9, 12)],
            (8, 5, 7),
            (3, 2, 4),
        ],
        [
            [(2, 1, 32, 32, 32), (1, 1, 19, 37, 23), (3, 1, 14, 9, 12)],
            (8, 5, 7),
            (0, 0, 0),
        ],
    ],
)
def test_whole_image_covered(
    data_shapes: Sequence[Sequence[int]],
    patch_size: Sequence[int],
    overlap: Sequence[int],
):
    patching_strategy = SequentialPatchingStrategy(data_shapes, patch_size, overlap)
    patch_specs = patching_strategy.patch_specs

    # track where patches have been sampled from
    tracking_arrays = [np.zeros(data_shape, dtype=bool) for data_shape in data_shapes]
    for patch_spec in patch_specs:
        tracking_array = tracking_arrays[patch_spec["data_idx"]]
        spatial_slice = tuple(
            slice(c, c + ps)
            for c, ps in zip(patch_spec["coords"], patch_spec["patch_size"])
        )
        # set to true where the patches would be sampled from
        tracking_array[(patch_spec["sample_idx"], slice(None), *spatial_slice)] = True

    for tracking_array in tracking_arrays:
        # if the patch specs covered all the image all the values should be true
        assert tracking_array.all()


# TODO: what else to test?



"""Write tiff function."""

from fnmatch import fnmatch
from pathlib import Path

import tifffile
from numpy.typing import NDArray

from careamics.config.support import SupportedData


def write_tiff(file_path: Path, img: NDArray, *args, **kwargs) -> None:
    # TODO: add link to tiffile docs for args kwrgs?
    """
    Write tiff files.

    Parameters
    ----------
    file_path : pathlib.Path
        Path to file.
    img : numpy.ndarray
        Image data to save.
    *args
        Positional arguments passed to `tifffile.imwrite`.
    **kwargs
        Keyword arguments passed to `tifffile.imwrite`.

    Raises
    ------
    ValueError
        When the file extension of `file_path` does not match the Unix shell-style
        pattern '*.tif*'.
    """
    if not fnmatch(
        file_path.suffix, SupportedData.get_extension_pattern(SupportedData.TIFF)
    ):
        raise ValueError(
            f"Unexpected extension '{file_path.suffix}' for save file type 'tiff'."
        )
    tifffile.imwrite(file_path, img, *args, **kwargs)



"""Transforms supported by CAREamics."""

from careamics.utils import BaseEnum


class SupportedTransform(str, BaseEnum):
    """Transforms officially supported by CAREamics."""

    XY_FLIP = "XYFlip"
    XY_RANDOM_ROTATE90 = "XYRandomRotate90"
    NORMALIZE = "Normalize"
    N2V_MANIPULATE = "N2VManipulate"



"""Function to export to the BioImage Model Zoo format."""

import tempfile
from pathlib import Path
from typing import Optional, Union

import numpy as np
from bioimageio.core import load_model_description, test_model
from bioimageio.spec import ValidationSummary, save_bioimageio_package
from pydantic import HttpUrl
from torch import __version__ as PYTORCH_VERSION
from torch import load, save
from torchvision import __version__ as TORCHVISION_VERSION

from careamics.config import Configuration, load_configuration, save_configuration
from careamics.config.support import SupportedArchitecture
from careamics.lightning.lightning_module import FCNModule, VAEModule
from careamics.utils.version import get_careamics_version

from .bioimage import (
    create_env_text,
    create_model_description,
    extract_model_path,
)
from .bioimage.cover_factory import create_cover


def _export_state_dict(
    model: Union[FCNModule, VAEModule], path: Union[Path, str]
) -> Path:
    """
    Export the model state dictionary to a file.

    Parameters
    ----------
    model : CAREamicsKiln
        CAREamics model to export.
    path : Union[Path, str]
        Path to the file where to save the model state dictionary.

    Returns
    -------
    Path
        Path to the saved model state dictionary.
    """
    path = Path(path)

    # make sure it has the correct suffix
    if path.suffix not in ".pth":
        path = path.with_suffix(".pth")

    # save model state dictionary
    # we save through the torch model itself to avoid the initial "model." in the
    # layers naming, which is incompatible with the way the BMZ load torch state dicts
    save(model.model.state_dict(), path)

    return path


def _load_state_dict(
    model: Union[FCNModule, VAEModule], path: Union[Path, str]
) -> None:
    """
    Load a model from a state dictionary.

    Parameters
    ----------
    model : CAREamicsKiln
        CAREamics model to be updated with the weights.
    path : Union[Path, str]
        Path to the model state dictionary.
    """
    path = Path(path)

    # load model state dictionary
    # same as in _export_state_dict, we load through the torch model to be compatible
    # witht bioimageio.core expectations for a torch state dict
    state_dict = load(path)
    model.model.load_state_dict(state_dict)


# TODO break down in subfunctions
def export_to_bmz(
    model: Union[FCNModule, VAEModule],
    config: Configuration,
    path_to_archive: Union[Path, str],
    model_name: str,
    general_description: str,
    data_description: str,
    authors: list[dict],
    input_array: np.ndarray,
    output_array: np.ndarray,
    covers: Optional[list[Union[Path, str]]] = None,
    channel_names: Optional[list[str]] = None,
    model_version: str = "0.1.0",
) -> None:
    """Export the model to BioImage Model Zoo format.

    Arrays are expected to be SC(Z)YX with singleton dimensions allowed for S and C.

    `model_name` should consist of letters, numbers, dashes, underscores and parentheses
    only.

    Parameters
    ----------
    model : CAREamicsModule
        CAREamics model to export.
    config : Configuration
        Model configuration.
    path_to_archive : Union[Path, str]
        Path to the output file.
    model_name : str
        Model name.
    general_description : str
        General description of the model.
    data_description : str
        Description of the data the model was trained on.
    authors : list[dict]
        Authors of the model.
    input_array : np.ndarray
        Input array, should not have been normalized.
    output_array : np.ndarray
        Output array, should have been denormalized.
    covers : list of pathlib.Path or str, default=None
        Paths to the cover images.
    channel_names : Optional[list[str]], optional
        Channel names, by default None.
    model_version : str, default="0.1.0"
        Model version.
    """
    path_to_archive = Path(path_to_archive)

    if path_to_archive.suffix != ".zip":
        raise ValueError(
            f"Path to archive must point to a zip file, got {path_to_archive}."
        )

    if not path_to_archive.parent.exists():
        path_to_archive.parent.mkdir(parents=True, exist_ok=True)

    # versions
    careamics_version = get_careamics_version()

    # save files in temporary folder
    with tempfile.TemporaryDirectory() as tmpdirname:
        temp_path = Path(tmpdirname)

        # create environment file
        # TODO move in bioimage module
        env_path = temp_path / "environment.yml"
        env_path.write_text(create_env_text(PYTORCH_VERSION, TORCHVISION_VERSION))

        # export input and ouputs
        inputs = temp_path / "inputs.npy"
        np.save(inputs, input_array)
        outputs = temp_path / "outputs.npy"
        np.save(outputs, output_array)

        # export configuration
        config_path = save_configuration(config, temp_path / "careamics.yaml")

        # export model state dictionary
        weight_path = _export_state_dict(model, temp_path / "weights.pth")

        # export cover if necesary
        if covers is None:
            covers = [create_cover(temp_path, input_array, output_array)]

        # create model description
        model_description = create_model_description(
            config=config,
            name=model_name,
            general_description=general_description,
            data_description=data_description,
            authors=authors,
            inputs=inputs,
            outputs=outputs,
            weights_path=weight_path,
            torch_version=PYTORCH_VERSION,
            careamics_version=careamics_version,
            config_path=config_path,
            env_path=env_path,
            covers=covers,
            channel_names=channel_names,
            model_version=model_version,
        )

        # test model description
        test_kwargs = (
            model_description.config.get("bioimageio", {})
            .get("test_kwargs", {})
            .get("pytorch_state_dict", {})
        )
        summary: ValidationSummary = test_model(model_description, **test_kwargs)
        if summary.status == "failed":
            raise ValueError(f"Model description test failed: {summary}")

        # save bmz model
        save_bioimageio_package(model_description, output_path=path_to_archive)


def load_from_bmz(
    path: Union[Path, str, HttpUrl]
) -> tuple[Union[FCNModule, VAEModule], Configuration]:
    """Load a model from a BioImage Model Zoo archive.

    Parameters
    ----------
    path : Path, str or HttpUrl
        Path to the BioImage Model Zoo archive. A Http URL must point to a downloadable
        location.

    Returns
    -------
    FCNModel or VAEModel
        The loaded CAREamics model.
    Configuration
        The loaded CAREamics configuration.

    Raises
    ------
    ValueError
        If the path is not a zip file.
    """
    # load description, this creates an unzipped folder next to the archive
    model_desc = load_model_description(path)

    # extract paths
    weights_path, config_path = extract_model_path(model_desc)

    # load configuration
    config = load_configuration(config_path)

    # create careamics lightning module
    if config.algorithm_config.model.architecture == SupportedArchitecture.UNET:
        model = FCNModule(algorithm_config=config.algorithm_config)
    elif config.algorithm_config.model.architecture == SupportedArchitecture.LVAE:
        model = VAEModule(algorithm_config=config.algorithm_config)
    else:
        raise ValueError(
            f"Unsupported architecture {config.algorithm_config.model.architecture}"
        )  # TODO ugly ?

    # load model state dictionary
    _load_state_dict(model, weights_path)

    return model, config



"""A package for the `PredictionWriterCallback` class and utilities."""

__all__ = [
    "CacheTiles",
    "PredictionWriterCallback",
    "WriteImage",
    "WriteStrategy",
    "WriteTilesZarr",
    "create_write_strategy",
    "select_write_extension",
    "select_write_func",
]

from .prediction_writer_callback import PredictionWriterCallback
from .write_strategy import CacheTiles, WriteImage, WriteStrategy, WriteTilesZarr
from .write_strategy_factory import (
    create_write_strategy,
    select_write_extension,
    select_write_func,
)



import pytest

from careamics.config.algorithms import CAREAlgorithm


def test_instantiation():
    """Test the instantiation of the CAREAlgorithm class."""
    model = {
        "architecture": "UNet",
    }
    CAREAlgorithm(model=model)


def test_no_n2v2():
    """Check that an error is raised if the model is set for n2v2."""
    model = {
        "architecture": "UNet",
        "n2v2": True,
    }

    with pytest.raises(ValueError):
        CAREAlgorithm(model=model)


def test_no_final_activation(minimum_algorithm_supervised: dict):
    """Check that an error is raised if the model has a final activation."""
    minimum_algorithm_supervised["model"] = {
        "architecture": "UNet",
        "final_activation": "ReLU",
    }
    with pytest.raises(ValueError):
        CAREAlgorithm(**minimum_algorithm_supervised)



"""Patch validation functions."""

from typing import Union

import numpy as np


def validate_patch_dimensions(
    arr: np.ndarray,
    patch_size: Union[list[int], tuple[int, ...]],
    is_3d_patch: bool,
) -> None:
    """
    Check patch size and array compatibility.

    This method validates the patch sizes with respect to the array dimensions:

    - Patch must have two dimensions fewer than the array (S and C).
    - Patch sizes are smaller than the corresponding array dimensions.

    If one of these conditions is not met, a ValueError is raised.

    This method should be called after inputs have been resized.

    Parameters
    ----------
    arr : np.ndarray
        Input array.
    patch_size : Union[list[int], tuple[int, ...]]
        Size of the patches along each dimension of the array, except the first.
    is_3d_patch : bool
        Whether the patch is 3D or not.

    Raises
    ------
    ValueError
        If the patch size is not consistent with the array shape (one more array
        dimension).
    ValueError
        If the patch size in Z is larger than the array dimension.
    ValueError
        If either of the patch sizes in X or Y is larger than the corresponding array
        dimension.
    """
    if len(patch_size) != len(arr.shape[2:]):
        raise ValueError(
            f"There must be a patch size for each spatial dimensions "
            f"(got {patch_size} patches for dims {arr.shape}). Check the axes order."
        )

    # Sanity checks on patch sizes versus array dimension
    if is_3d_patch and patch_size[0] > arr.shape[-3]:
        raise ValueError(
            f"Z patch size is inconsistent with image shape "
            f"(got {patch_size[0]} patches for dim {arr.shape[1]}). Check the axes "
            f"order."
        )

    if patch_size[-2] > arr.shape[-2] or patch_size[-1] > arr.shape[-1]:
        raise ValueError(
            f"At least one of YX patch dimensions is larger than the corresponding "
            f"image dimension (got {patch_size} patches for dims {arr.shape[-2:]}). "
            f"Check the axes order."
        )



"""Function to iterate over files."""

from __future__ import annotations

from collections.abc import Generator
from pathlib import Path
from typing import Callable, Optional, Union

from numpy.typing import NDArray
from torch.utils.data import get_worker_info

from careamics.config import DataConfig, InferenceConfig
from careamics.file_io.read import read_tiff
from careamics.utils.logging import get_logger

from .dataset_utils import reshape_array

logger = get_logger(__name__)


def iterate_over_files(
    data_config: Union[DataConfig, InferenceConfig],
    data_files: list[Path],
    target_files: Optional[list[Path]] = None,
    read_source_func: Callable = read_tiff,
) -> Generator[tuple[NDArray, Optional[NDArray]], None, None]:
    """Iterate over data source and yield whole reshaped images.

    Parameters
    ----------
    data_config : CAREamics DataConfig or InferenceConfig
        Configuration.
    data_files : list of pathlib.Path
        List of data files.
    target_files : list of pathlib.Path, optional
        List of target files, by default None.
    read_source_func : Callable, optional
        Function to read the source, by default read_tiff.

    Yields
    ------
    NDArray
        Image.
    """
    # When num_workers > 0, each worker process will have a different copy of the
    # dataset object
    # Configuring each copy independently to avoid having duplicate data returned
    # from the workers
    worker_info = get_worker_info()
    worker_id = worker_info.id if worker_info is not None else 0
    num_workers = worker_info.num_workers if worker_info is not None else 1

    # iterate over the files
    for i, filename in enumerate(data_files):
        # retrieve file corresponding to the worker id
        if i % num_workers == worker_id:
            try:
                # read data
                sample = read_source_func(filename, data_config.axes)

                # reshape array
                reshaped_sample = reshape_array(sample, data_config.axes)

                # read target, if available
                if target_files is not None:
                    if filename.name != target_files[i].name:
                        raise ValueError(
                            f"File {filename} does not match target file "
                            f"{target_files[i]}. Have you passed sorted "
                            f"arrays?"
                        )

                    # read target
                    target = read_source_func(target_files[i], data_config.axes)

                    # reshape target
                    reshaped_target = reshape_array(target, data_config.axes)

                    yield reshaped_sample, reshaped_target
                else:
                    yield reshaped_sample, None

            except Exception as e:
                logger.error(f"Error reading file {filename}: {e}")



from careamics.config.architectures import UNetModel
from careamics.models import model_factory
from careamics.models.unet import UNet


# TODO improve tests
def test_model_registry_unet():
    """Test that"""
    model_config = {
        "architecture": "UNet",
    }

    # instantiate model
    model = model_factory(UNetModel(**model_config))
    assert isinstance(model, UNet)



from careamics.utils.version import get_careamics_version


def test_get_clean_version():
    """Test that the result has three members (major, minor, patch) and that the last
    one does not include `dev`."""
    version_members = get_careamics_version().split(".")

    assert len(version_members) == 3
    assert "dev" not in version_members[-1]



"""N2V manipulation transform for PyTorch."""

import platform
from typing import Any, Optional

import torch

from careamics.config.support import SupportedPixelManipulation, SupportedStructAxis
from careamics.config.transformations import N2VManipulateModel

from .pixel_manipulation_torch import (
    median_manipulate_torch,
    uniform_manipulate_torch,
)
from .struct_mask_parameters import StructMaskParameters


class N2VManipulateTorch:
    """
    Default augmentation for the N2V model.

    This transform expects C(Z)YX dimensions.

    Parameters
    ----------
    n2v_manipulate_config : N2VManipulateConfig
        N2V manipulation configuration.
    seed : Optional[int], optional
        Random seed, by default None.

    Attributes
    ----------
    masked_pixel_percentage : float
        Percentage of pixels to mask.
    roi_size : int
        Size of the replacement area.
    strategy : Literal[ "uniform", "median" ]
        Replacement strategy, uniform or median.
    remove_center : bool
        Whether to remove central pixel from patch.
    struct_mask : Optional[StructMaskParameters]
        StructN2V mask parameters.
    rng : Generator
        Random number generator.
    """

    def __init__(
        self,
        n2v_manipulate_config: N2VManipulateModel,
        seed: Optional[int] = None,
    ):
        """Constructor.

        Parameters
        ----------
        n2v_manipulate_config : N2VManipulateModel
            N2V manipulation configuration.
        seed : Optional[int], optional
            Random seed, by default None.
        """
        self.masked_pixel_percentage = n2v_manipulate_config.masked_pixel_percentage
        self.roi_size = n2v_manipulate_config.roi_size
        self.strategy = n2v_manipulate_config.strategy
        self.remove_center = n2v_manipulate_config.remove_center

        if n2v_manipulate_config.struct_mask_axis == SupportedStructAxis.NONE:
            self.struct_mask: Optional[StructMaskParameters] = None
        else:
            self.struct_mask = StructMaskParameters(
                axis=(
                    0
                    if n2v_manipulate_config.struct_mask_axis
                    == SupportedStructAxis.HORIZONTAL
                    else 1
                ),
                span=n2v_manipulate_config.struct_mask_span,
            )

        # PyTorch random generator
        # TODO refactor into careamics.utils.torch_utils.get_device
        if torch.cuda.is_available():
            device = "cuda"
        elif torch.backends.mps.is_available() and platform.processor() in (
            "arm",
            "arm64",
        ):
            device = "mps"
        else:
            device = "cpu"

        self.rng = (
            torch.Generator(device=device).manual_seed(seed)
            if seed is not None
            else torch.Generator(device=device)
        )

    def __call__(
        self, batch: torch.Tensor, *args: Any, **kwargs: Any
    ) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """Apply the transform to the image.

        Parameters
        ----------
        batch : torch.Tensor
            Batch if image patches, 2D or 3D, shape BC(Z)YX.
        *args : Any
            Additional arguments, unused.
        **kwargs : Any
            Additional keyword arguments, unused.

        Returns
        -------
        tuple[torch.Tensor, torch.Tensor, torch.Tensor]
            Masked patch, original patch, and mask.
        """
        masked = torch.zeros_like(batch)
        mask = torch.zeros_like(batch, dtype=torch.uint8)

        if self.strategy == SupportedPixelManipulation.UNIFORM:
            # Iterate over the channels to apply manipulation separately
            for c in range(batch.shape[1]):
                masked[:, c, ...], mask[:, c, ...] = uniform_manipulate_torch(
                    patch=batch[:, c, ...],
                    mask_pixel_percentage=self.masked_pixel_percentage,
                    subpatch_size=self.roi_size,
                    remove_center=self.remove_center,
                    struct_params=self.struct_mask,
                    rng=self.rng,
                )
        elif self.strategy == SupportedPixelManipulation.MEDIAN:
            # Iterate over the channels to apply manipulation separately
            for c in range(batch.shape[1]):
                masked[:, c, ...], mask[:, c, ...] = median_manipulate_torch(
                    batch=batch[:, c, ...],
                    mask_pixel_percentage=self.masked_pixel_percentage,
                    subpatch_size=self.roi_size,
                    struct_params=self.struct_mask,
                    rng=self.rng,
                )
        else:
            raise ValueError(f"Unknown masking strategy ({self.strategy}).")

        return masked, batch, mask



"""Zarr dataset."""

# from itertools import islice
# from typing import Callable, Dict, List, Optional, Tuple, Union

# import numpy as np
# import torch
# import zarr

# from careamics.utils import RunningStats
# from careamics.utils.logging import get_logger

# from ..utils import normalize
# from .dataset_utils import read_zarr
# from .patching.patching import (
#     generate_patches_unsupervised,
# )

# logger = get_logger(__name__)


# class ZarrDataset(torch.utils.data.IterableDataset):
#     """Dataset to extract patches from a zarr storage.

#     Parameters
#     ----------
#     data_source : Union[zarr.Group, zarr.Array]
#         Zarr storage.
#     axes : str
#         Description of axes in format STCZYX.
#     patch_extraction_method : Union[ExtractionStrategies, None]
#         Patch extraction strategy, as defined in extraction_strategy.
#     patch_size : Optional[Union[List[int], Tuple[int]]], optional
#         Size of the patches in each dimension, by default None.
#     num_patches : Optional[int], optional
#         Number of patches to extract, by default None.
#     mean : Optional[float], optional
#         Expected mean of the dataset, by default None.
#     std : Optional[float], optional
#         Expected standard deviation of the dataset, by default None.
#     patch_transform : Optional[Callable], optional
#         Patch transform callable, by default None.
#     patch_transform_params : Optional[Dict], optional
#         Patch transform parameters, by default None.
#     running_stats_window_perc : float, optional
#         Percentage of the dataset to use for calculating the initial mean and standard
#         deviation, by default 0.01.
#     mode : str, optional
#         train/predict, controls running stats calculation.
#     """

#     def __init__(
#         self,
#         data_source: Union[zarr.Group, zarr.Array],
#         axes: str,
#         patch_extraction_method: Union[SupportedExtractionStrategy, None],
#         patch_size: Optional[Union[List[int], Tuple[int]]] = None,
#         num_patches: Optional[int] = None,
#         mean: Optional[float] = None,
#         std: Optional[float] = None,
#         patch_transform: Optional[Callable] = None,
#         patch_transform_params: Optional[Dict] = None,
#         running_stats_window_perc: float = 0.01,
#         mode: str = "train",
#     ) -> None:
#         self.data_source = data_source
#         self.axes = axes
#         self.patch_extraction_method = patch_extraction_method
#         self.patch_size = patch_size
#         self.num_patches = num_patches
#         self.mean = mean
#         self.std = std
#         self.patch_transform = patch_transform
#         self.patch_transform_params = patch_transform_params
#         self.sample = read_zarr(self.data_source, self.axes)
#         self.running_stats_window = int(
#             np.prod(self.sample._cdata_shape) * running_stats_window_perc
#         )
#         self.mode = mode
#         self.running_stats = RunningStats()

#         self._calculate_initial_mean_std()

#     def _calculate_initial_mean_std(self):
#         """Calculate initial mean and std of the dataset."""
#         if self.mean is None and self.std is None:
#             idxs = np.random.randint(
#                 0,
#                 np.prod(self.sample._cdata_shape),
#                 size=max(1, self.running_stats_window),
#             )
#             random_chunks = self.sample[idxs]
#             self.running_stats.init(random_chunks.mean(), random_chunks.std())

#     def _generate_patches(self):
#         """Generate patches from the dataset and calculates running stats.

#         Yields
#         ------
#         np.ndarray
#             Patch.
#         """
#         patches = generate_patches_unsupervised(
#             self.sample,
#             self.patch_extraction_method,
#             self.patch_size,
#         )

#         # num_patches = np.ceil(
#         #     np.prod(self.sample.chunks)
#         #     / (np.prod(self.patch_size) * self.running_stats_window)
#         # ).astype(int)

#         for idx, patch in enumerate(patches):
#             if self.mode != "predict":
#                 self.running_stats.update(patch.mean())
#             if isinstance(patch, tuple):
#                 normalized_patch = normalize(
#                     img=patch[0],
#                     mean=self.running_stats.avg_mean.value,
#                     std=self.running_stats.avg_std.value,
#                 )
#                 patch = (normalized_patch, *patch[1:])
#             else:
#                 patch = normalize(
#                     img=patch,
#                     mean=self.running_stats.avg_mean.value,
#                     std=self.running_stats.avg_std.value,
#                 )

#             if self.patch_transform is not None:
#                 assert self.patch_transform_params is not None
#                 patch = self.patch_transform(patch, **self.patch_transform_params)
#             if self.num_patches is not None and idx >= self.num_patches:
#                 return
#             else:
#                 yield patch
#         self.mean = self.running_stats.avg_mean.value
#         self.std = self.running_stats.avg_std.value

#     def __iter__(self):
#         """
#         Iterate over data source and yield single patch.

#         Yields
#         ------
#         np.ndarray
#         """
#         worker_info = torch.utils.data.get_worker_info()
#         num_workers = worker_info.num_workers if worker_info is not None else 1
#         yield from islice(self._generate_patches(), 0, None, num_workers)



from pathlib import Path
from typing import Callable

import numpy as np
import pytest

from careamics import CAREamist, Configuration
from careamics.config.support import SupportedData
from careamics.model_io import export_to_bmz


@pytest.fixture
def gaussian_likelihood_params():
    return {"predict_logvar": "pixelwise", "logvar_lowerbound": -5}


# TODO add details about where each of these fixture is used (e.g. smoke test)
@pytest.fixture
def create_tiff(path: Path, n_files: int):
    """Create tiff files for testing."""
    if not path.exists():
        path.mkdir()

    for i in range(n_files):
        file_path = path / f"file_{i}.tif"
        file_path.touch()


@pytest.fixture
def minimum_algorithm_n2v() -> dict:
    """Create a minimum algorithm dictionary.

    Returns
    -------
    dict
        A minimum algorithm example.
    """
    # create dictionary
    algorithm = {
        "algorithm": "n2v",
        "loss": "n2v",
        "model": {
            "architecture": "UNet",
        },
    }

    return algorithm


@pytest.fixture
def minimum_algorithm_supervised() -> dict:
    """Create a minimum algorithm dictionary.

    Returns
    -------
    dict
        A minimum algorithm example.
    """
    # create dictionary
    algorithm = {
        "algorithm": "n2n",
        "loss": "mae",
        "model": {
            "architecture": "UNet",
        },
    }

    return algorithm


# TODO: wrong! need to update/remove this fixture
@pytest.fixture
def minimum_algorithm_musplit() -> dict:
    """Create a minimum algorithm dictionary.

    Returns
    -------
    dict
        A minimum algorithm example.
    """
    # create dictionary
    algorithm = {
        "algorithm_type": "vae",
        "algorithm": "musplit",  # TODO temporary
        "loss": "musplit",
        "model": {
            "architecture": "LVAE",
            "z_dims": (128, 128, 128),
            "multiscale_count": 2,
            "predict_logvar": "pixelwise",
        },
        "likelihood": {
            "type": "GaussianLikelihoodConfig",
        },
    }

    return algorithm


# TODO: wrong! need to update/remove this fixture
@pytest.fixture
def minimum_algorithm_denoisplit() -> dict:
    """Create a minimum algorithm dictionary.

    Returns
    -------
    dict
        A minimum algorithm example.
    """
    # create dictionary
    algorithm = {
        "algorithm_type": "vae",
        "algorithm": "denoisplit",
        "loss": "denoisplit",
        "model": {
            "architecture": "LVAE",
            "z_dims": (128, 128, 128),
            "multiscale_count": 2,
        },
        "likelihood": {"type": "GaussianLikelihoodConfig", "color_channels": 2},
        "noise_model": "MultiChannelNMConfig",
    }

    return algorithm


@pytest.fixture
def minimum_data() -> dict:
    """Create a minimum data dictionary.

    Returns
    -------
    dict
        A minimum data example.
    """
    # create dictionary
    data = {
        "data_type": SupportedData.ARRAY.value,
        "patch_size": [8, 8],
        "axes": "YX",
    }

    return data


@pytest.fixture
def minimum_inference() -> dict:
    """Create a minimum inference dictionary.

    Returns
    -------
    dict
        A minimum data example.
    """
    # create dictionary
    predic = {
        "data_type": SupportedData.ARRAY.value,
        "axes": "YX",
        "image_means": [2.0],
        "image_stds": [1.0],
    }

    return predic


@pytest.fixture
def minimum_training() -> dict:
    """Create a minimum training dictionary.

    Returns
    -------
    dict
        A minimum training example.
    """
    # create dictionary
    training = {
        "num_epochs": 1,
    }

    return training


@pytest.fixture
def minimum_n2v_configuration(
    minimum_algorithm_n2v: dict, minimum_data: dict, minimum_training: dict
) -> dict:
    """Create a minimum configuration dictionary.

    Parameters
    ----------
    tmp_path : Path
        Temporary path for testing.
    minimum_algorithm : dict
        Minimum algorithm configuration.
    minimum_data_n2v : dict
        Minimum N2V data configuration.
    minimum_training : dict
        Minimum training configuration.

    Returns
    -------
    dict
        A minumum configuration example.
    """
    # create dictionary
    configuration = {
        "experiment_name": "LevitatingFrog",
        "algorithm_config": minimum_algorithm_n2v,
        "training_config": minimum_training,
        "data_config": minimum_data,
    }

    return configuration


@pytest.fixture
def minimum_supervised_configuration(
    minimum_algorithm_supervised: dict, minimum_data: dict, minimum_training: dict
) -> dict:
    configuration = {
        "experiment_name": "LevitatingFrog",
        "algorithm_config": minimum_algorithm_supervised,
        "training_config": minimum_training,
        "data_config": minimum_data,
    }

    return configuration


@pytest.fixture
def ordered_array() -> Callable:
    """A function that returns an array with ordered values."""

    def _ordered_array(shape: tuple, dtype=int) -> np.ndarray:
        """An array with ordered values.

        Parameters
        ----------
        shape : tuple
            Shape of the array.

        Returns
        -------
        np.ndarray
            Array with ordered values.
        """
        return np.arange(np.prod(shape), dtype=dtype).reshape(shape).astype(np.float32)

    return _ordered_array


@pytest.fixture
def array_2D() -> np.ndarray:
    """A 2D array with shape (1, 3, 10, 9).

    Returns
    -------
    np.ndarray
        2D array with shape (1, 3, 10, 9).
    """
    return np.arange(90 * 3).reshape((1, 3, 10, 9))


@pytest.fixture
def array_3D() -> np.ndarray:
    """A 3D array with shape (1, 3, 5, 10, 9).

    Returns
    -------
    np.ndarray
        3D array with shape (1, 3, 5, 10, 9).
    """
    return np.arange(2048 * 3).reshape((1, 3, 8, 16, 16))


@pytest.fixture
def patch_size() -> tuple[int, int]:
    return (64, 64)


@pytest.fixture
def overlaps() -> tuple[int, int]:
    return (32, 32)


@pytest.fixture
def pre_trained(tmp_path, minimum_n2v_configuration):
    """Fixture to create a pre-trained CAREamics model."""
    # training data
    train_array = np.arange(32 * 32).reshape((32, 32)).astype(np.float32)

    # create configuration
    config = Configuration(**minimum_n2v_configuration)
    config.training_config.num_epochs = 1
    config.data_config.axes = "YX"
    config.data_config.batch_size = 2
    config.data_config.data_type = SupportedData.ARRAY.value
    config.data_config.patch_size = (8, 8)

    # instantiate CAREamist
    careamist = CAREamist(source=config, work_dir=tmp_path)

    # train CAREamist
    careamist.train(train_source=train_array)

    # check that it trained
    pre_trained_path: Path = tmp_path / "checkpoints" / "last.ckpt"
    assert pre_trained_path.exists()

    return pre_trained_path


@pytest.fixture
def pre_trained_bmz(tmp_path, pre_trained) -> Path:
    """Fixture to create a BMZ model."""
    # training data
    train_array = np.ones((32, 32), dtype=np.float32)

    # instantiate CAREamist
    careamist = CAREamist(source=pre_trained, work_dir=tmp_path)

    # predict (no tiling and no tta)
    predicted_output = careamist.predict(train_array, tta_transforms=False)
    predicted = np.concatenate(predicted_output, axis=0)

    # export to BioImage Model Zoo
    path = tmp_path / "model.zip"
    export_to_bmz(
        model=careamist.model,
        config=careamist.cfg,
        path_to_archive=path,
        model_name="TopModel",
        general_description="A model that just walked in.",
        data_description="My data.",
        authors=[{"name": "Amod", "affiliation": "El"}],
        input_array=train_array[np.newaxis, np.newaxis, ...],
        output_array=predicted,
    )
    assert path.exists()

    return path


@pytest.fixture
def create_dummy_noise_model(
    n_gaussians: int = 3,
    n_coeffs: int = 3,
) -> None:
    weights = np.random.rand(3 * n_gaussians, n_coeffs)
    nm_dict = {
        "trained_weight": weights,
        "min_signal": np.array([0]),
        "max_signal": np.array([2**16 - 1]),
        "min_sigma": 0.125,
    }
    return nm_dict


@pytest.fixture
def minimum_lvae_params():
    return {
        "input_shape": (64, 64),
        "output_channels": 2,
        "multiscale_count": 1,
        "encoder_conv_strides": [2, 2],
        "decoder_conv_strides": [2, 2],
        "z_dims": [128, 128, 128, 128],
        "encoder_n_filters": 64,
        "decoder_n_filters": 64,
        "encoder_dropout": 0.1,
        "decoder_dropout": 0.1,
        "nonlinearity": "ELU",
        "predict_logvar": "pixelwise",
        "analytical_kl": False,
    }



"""I/O functions for Configuration objects."""

from pathlib import Path
from typing import Union

import yaml

from careamics.config import Configuration


def load_configuration(path: Union[str, Path]) -> Configuration:
    """
    Load configuration from a yaml file.

    Parameters
    ----------
    path : str or Path
        Path to the configuration.

    Returns
    -------
    Configuration
        Configuration.

    Raises
    ------
    FileNotFoundError
        If the configuration file does not exist.
    """
    # load dictionary from yaml
    if not Path(path).exists():
        raise FileNotFoundError(
            f"Configuration file {path} does not exist in " f" {Path.cwd()!s}"
        )

    dictionary = yaml.load(Path(path).open("r"), Loader=yaml.SafeLoader)

    return Configuration(**dictionary)


def save_configuration(config: Configuration, path: Union[str, Path]) -> Path:
    """
    Save configuration to path.

    Parameters
    ----------
    config : Configuration
        Configuration to save.
    path : str or Path
        Path to a existing folder in which to save the configuration, or to a valid
        configuration file path (uses a .yml or .yaml extension).

    Returns
    -------
    Path
        Path object representing the configuration.

    Raises
    ------
    ValueError
        If the path does not point to an existing directory or .yml file.
    """
    # make sure path is a Path object
    config_path = Path(path)

    # check if path is pointing to an existing directory or .yml file
    if config_path.exists():
        if config_path.is_dir():
            config_path = Path(config_path, "config.yml")
        elif config_path.suffix != ".yml" and config_path.suffix != ".yaml":
            raise ValueError(
                f"Path must be a directory or .yml or .yaml file (got {config_path})."
            )
    else:
        if config_path.suffix != ".yml" and config_path.suffix != ".yaml":
            raise ValueError(
                f"Path must be a directory or .yml or .yaml file (got {config_path})."
            )

    # save configuration as dictionary to yaml
    with open(config_path, "w") as f:
        # dump configuration
        yaml.dump(config.model_dump(), f, default_flow_style=False, sort_keys=False)

    return config_path



"""In-memory dataset module."""

from __future__ import annotations

import copy
from pathlib import Path
from typing import Any, Callable, Optional, Union

import numpy as np
from torch.utils.data import Dataset

from careamics.config import DataConfig
from careamics.config.transformations import NormalizeModel
from careamics.dataset.patching.patching import (
    PatchedOutput,
    Stats,
    prepare_patches_supervised,
    prepare_patches_supervised_array,
    prepare_patches_unsupervised,
    prepare_patches_unsupervised_array,
)
from careamics.file_io.read import read_tiff
from careamics.transforms import Compose
from careamics.utils.logging import get_logger

logger = get_logger(__name__)


class InMemoryDataset(Dataset):
    """Dataset storing data in memory and allowing generating patches from it.

    Parameters
    ----------
    data_config : CAREamics DataConfig
        (see careamics.config.data_model.DataConfig)
        Data configuration.
    inputs : numpy.ndarray or list[pathlib.Path]
        Input data.
    input_target : numpy.ndarray or list[pathlib.Path], optional
        Target data, by default None.
    read_source_func : Callable, optional
        Read source function for custom types, by default read_tiff.
    **kwargs : Any
        Additional keyword arguments, unused.
    """

    def __init__(
        self,
        data_config: DataConfig,
        inputs: Union[np.ndarray, list[Path]],
        input_target: Optional[Union[np.ndarray, list[Path]]] = None,
        read_source_func: Callable = read_tiff,
        **kwargs: Any,
    ) -> None:
        """
        Constructor.

        Parameters
        ----------
        data_config : GeneralDataConfig
            Data configuration.
        inputs : numpy.ndarray or list[pathlib.Path]
            Input data.
        input_target : numpy.ndarray or list[pathlib.Path], optional
            Target data, by default None.
        read_source_func : Callable, optional
            Read source function for custom types, by default read_tiff.
        **kwargs : Any
            Additional keyword arguments, unused.
        """
        self.data_config = data_config
        self.inputs = inputs
        self.input_targets = input_target
        self.axes = self.data_config.axes
        self.patch_size = self.data_config.patch_size

        # read function
        self.read_source_func = read_source_func

        # generate patches
        supervised = self.input_targets is not None
        patches_data = self._prepare_patches(supervised)

        # unpack the dataclass
        self.data = patches_data.patches
        self.data_targets = patches_data.targets

        # set image statistics
        if self.data_config.image_means is None:
            self.image_stats = patches_data.image_stats
            logger.info(
                f"Computed dataset mean: {self.image_stats.means}, "
                f"std: {self.image_stats.stds}"
            )
        else:
            self.image_stats = Stats(
                self.data_config.image_means, self.data_config.image_stds
            )

        # set target statistics
        if self.data_config.target_means is None:
            self.target_stats = patches_data.target_stats
        else:
            self.target_stats = Stats(
                self.data_config.target_means, self.data_config.target_stds
            )

        # update mean and std in configuration
        # the object is mutable and should then be recorded in the CAREamist obj
        self.data_config.set_means_and_stds(
            image_means=self.image_stats.means,
            image_stds=self.image_stats.stds,
            target_means=self.target_stats.means,
            target_stds=self.target_stats.stds,
        )
        # get transforms
        self.patch_transform = Compose(
            transform_list=[
                NormalizeModel(
                    image_means=self.image_stats.means,
                    image_stds=self.image_stats.stds,
                    target_means=self.target_stats.means,
                    target_stds=self.target_stats.stds,
                )
            ]
            + list(self.data_config.transforms),
        )

    def _prepare_patches(self, supervised: bool) -> PatchedOutput:
        """
        Iterate over data source and create an array of patches.

        Parameters
        ----------
        supervised : bool
            Whether the dataset is supervised or not.

        Returns
        -------
        numpy.ndarray
            Array of patches.
        """
        if supervised:
            if isinstance(self.inputs, np.ndarray) and isinstance(
                self.input_targets, np.ndarray
            ):
                return prepare_patches_supervised_array(
                    self.inputs,
                    self.axes,
                    self.input_targets,
                    self.patch_size,
                )
            elif isinstance(self.inputs, list) and isinstance(self.input_targets, list):
                return prepare_patches_supervised(
                    self.inputs,
                    self.input_targets,
                    self.axes,
                    self.patch_size,
                    self.read_source_func,
                )
            else:
                raise ValueError(
                    f"Data and target must be of the same type, either both numpy "
                    f"arrays or both lists of paths, got {type(self.inputs)} (data) "
                    f"and {type(self.input_targets)} (target)."
                )
        else:
            if isinstance(self.inputs, np.ndarray):
                return prepare_patches_unsupervised_array(
                    self.inputs,
                    self.axes,
                    self.patch_size,
                )
            else:
                return prepare_patches_unsupervised(
                    self.inputs,
                    self.axes,
                    self.patch_size,
                    self.read_source_func,
                )

    def __len__(self) -> int:
        """
        Return the length of the dataset.

        Returns
        -------
        int
            Length of the dataset.
        """
        return self.data.shape[0]

    def __getitem__(self, index: int) -> tuple[np.ndarray, ...]:
        """
        Return the patch corresponding to the provided index.

        Parameters
        ----------
        index : int
            Index of the patch to return.

        Returns
        -------
        tuple of numpy.ndarray
            Patch.

        Raises
        ------
        ValueError
            If dataset mean and std are not set.
        """
        patch = self.data[index]

        # if there is a target
        if self.data_targets is not None:
            # get target
            target = self.data_targets[index]
            return self.patch_transform(patch=patch, target=target)

        return self.patch_transform(patch=patch)

    def get_data_statistics(self) -> tuple[list[float], list[float]]:
        """Return training data statistics.

        This does not return the target data statistics, only those of the input.

        Returns
        -------
        tuple of list of floats
            Means and standard deviations across channels of the training data.
        """
        return self.image_stats.get_statistics()

    def split_dataset(
        self,
        percentage: float = 0.1,
        minimum_patches: int = 1,
    ) -> InMemoryDataset:
        """Split a new dataset away from the current one.

        This method is used to extract random validation patches from the dataset.

        Parameters
        ----------
        percentage : float, optional
            Percentage of patches to extract, by default 0.1.
        minimum_patches : int, optional
            Minimum number of patches to extract, by default 5.

        Returns
        -------
        CAREamics InMemoryDataset
            New dataset with the extracted patches.

        Raises
        ------
        ValueError
            If `percentage` is not between 0 and 1.
        ValueError
            If `minimum_number` is not between 1 and the number of patches.
        """
        if percentage < 0 or percentage > 1:
            raise ValueError(f"Percentage must be between 0 and 1, got {percentage}.")

        if minimum_patches < 1 or minimum_patches > len(self):
            raise ValueError(
                f"Minimum number of patches must be between 1 and "
                f"{len(self)} (number of patches), got "
                f"{minimum_patches}. Adjust the patch size or the minimum number of "
                f"patches."
            )

        total_patches = len(self)

        # number of patches to extract (either percentage rounded or minimum number)
        n_patches = max(round(total_patches * percentage), minimum_patches)

        # get random indices
        indices = np.random.choice(total_patches, n_patches, replace=False)

        # extract patches
        val_patches = self.data[indices]

        # remove patches from self.patch
        self.data = np.delete(self.data, indices, axis=0)

        # same for targets
        if self.data_targets is not None:
            val_targets = self.data_targets[indices]
            self.data_targets = np.delete(self.data_targets, indices, axis=0)

        # clone the dataset
        dataset = copy.deepcopy(self)

        # reassign patches
        dataset.data = val_patches

        # reassign targets
        if self.data_targets is not None:
            dataset.data_targets = val_targets

        return dataset



"""Configuration building convenience functions for the CAREamics CLI."""

import sys
from dataclasses import dataclass
from pathlib import Path
from typing import Annotated, Optional

import click
import typer
import yaml

from ..config import (
    Configuration,
    create_care_configuration,
    create_n2n_configuration,
    create_n2v_configuration,
    save_configuration,
)
from .utils import handle_2D_3D_callback

WORK_DIR = Path.cwd()

app = typer.Typer()


def _config_builder_exit(ctx: typer.Context, config: Configuration) -> None:
    """
    Function to be called at the end of a CLI configuration builder.

    Saves the `config` object and performs other functionality depending on the command
    context.

    Parameters
    ----------
    ctx : typer.Context
        Typer Context.
    config : Configuration
        CAREamics configuration.
    """
    conf_path = (ctx.obj.dir / ctx.obj.name).with_suffix(".yaml")
    save_configuration(config, conf_path)
    if ctx.obj.print:
        print(yaml.dump(config.model_dump(), indent=2))


@dataclass
class ConfOptions:
    """Data class for containing CLI `conf` command option values."""

    dir: Path
    name: str
    force: bool
    print: bool


@app.callback()
def conf_options(  # numpydoc ignore=PR01
    ctx: typer.Context,
    dir: Annotated[
        Path,
        typer.Option(
            "--dir", "-d", exists=True, help="Directory to save the config file to."
        ),
    ] = WORK_DIR,
    name: Annotated[
        str, typer.Option("--name", "-n", help="The config file name.")
    ] = "config",
    force: Annotated[
        bool,
        typer.Option(
            "--force", "-f", help="Whether to overwrite existing config files."
        ),
    ] = False,
    print: Annotated[
        bool,
        typer.Option(
            "--print",
            "-p",
            help="Whether to print the config file to the console.",
        ),
    ] = False,
):
    """Build and save CAREamics configuration files."""
    # Callback is called still on --help command
    # If a config exists it will complain that you need to use the -f flag
    if "--help" in sys.argv:
        return
    conf_path = (dir / name).with_suffix(".yaml")
    if conf_path.exists() and not force:
        raise FileExistsError(f"To overwrite '{conf_path}' use flag --force/-f.")

    ctx.obj = ConfOptions(dir, name, force, print)


# TODO: Need to decide how to parse model kwargs
#   - Could be json style string to be loaded as dict e.g. {"depth": 3}
#        - Cons: Annoying to type, easily have syntax errors
#   - Could parse all unknown options as model kwargs
#       - Cons: There could be argument name clashes


@app.command()
def care(  # numpydoc ignore=PR01
    ctx: typer.Context,
    experiment_name: Annotated[str, typer.Option(help="Name of the experiment.")],
    axes: Annotated[str, typer.Option(help="Axes of the data (e.g. SYX).")],
    patch_size: Annotated[
        click.Tuple,
        typer.Option(
            help=(
                "Size of the patches along the spatial dimensions (if the data "
                "is not 3D pass the last value as -1 e.g. --patch-size 64 64 -1)."
            ),
            click_type=click.Tuple([int, int, int]),
            callback=handle_2D_3D_callback,
        ),
    ],
    batch_size: Annotated[int, typer.Option(help="Batch size.")],
    num_epochs: Annotated[int, typer.Option(help="Number of epochs.")],
    data_type: Annotated[
        click.Choice,
        typer.Option(click_type=click.Choice(["tiff"]), help="Type of the data."),
    ] = "tiff",
    use_augmentations: Annotated[
        bool, typer.Option(help="Whether to use augmentations.")
    ] = True,
    independent_channels: Annotated[
        bool, typer.Option(help="Whether to train all channels independently.")
    ] = False,
    loss: Annotated[
        click.Choice,
        typer.Option(
            click_type=click.Choice(["mae", "mse"]),
            help="Loss function to use.",
        ),
    ] = "mae",
    n_channels_in: Annotated[
        Optional[int], typer.Option(help="Number of channels in")
    ] = None,
    n_channels_out: Annotated[
        Optional[int], typer.Option(help="Number of channels out")
    ] = None,
    logger: Annotated[
        click.Choice,
        typer.Option(
            click_type=click.Choice(["wandb", "tensorboard", "none"]),
            help="Logger to use.",
        ),
    ] = "none",
    # TODO: How to address model kwargs
) -> None:
    """
    Create a configuration for training CARE.

    If "Z" is present in `axes`, then `path_size` must be a list of length 3, otherwise
    2.

    If "C" is present in `axes`, then you need to set `n_channels_in` to the number of
    channels. Likewise, if you set the number of channels, then "C" must be present in
    `axes`.

    To set the number of output channels, use the `n_channels_out` parameter. If it is
    not specified, it will be assumed to be equal to `n_channels_in`.

    By default, all channels are trained together. To train all channels independently,
    set `independent_channels` to True.

    By setting `use_augmentations` to False, the only transformation applied will be
    normalization.
    """
    config = create_care_configuration(
        experiment_name=experiment_name,
        data_type=data_type,
        axes=axes,
        patch_size=patch_size,
        batch_size=batch_size,
        num_epochs=num_epochs,
        # TODO: fix choosing augmentations
        augmentations=None if use_augmentations else [],
        independent_channels=independent_channels,
        loss=loss,
        n_channels_in=n_channels_in,
        n_channels_out=n_channels_out,
        logger=logger,
    )
    _config_builder_exit(ctx, config)


@app.command()
def n2n(  # numpydoc ignore=PR01
    ctx: typer.Context,
    experiment_name: Annotated[str, typer.Option(help="Name of the experiment.")],
    axes: Annotated[str, typer.Option(help="Axes of the data (e.g. SYX).")],
    patch_size: Annotated[
        click.Tuple,
        typer.Option(
            help=(
                "Size of the patches along the spatial dimensions (if the data "
                "is not 3D pass the last value as -1 e.g. --patch-size 64 64 -1)."
            ),
            click_type=click.Tuple([int, int, int]),
            callback=handle_2D_3D_callback,
        ),
    ],
    batch_size: Annotated[int, typer.Option(help="Batch size.")],
    num_epochs: Annotated[int, typer.Option(help="Number of epochs.")],
    data_type: Annotated[
        click.Choice,
        typer.Option(click_type=click.Choice(["tiff"]), help="Type of the data."),
    ] = "tiff",
    use_augmentations: Annotated[
        bool, typer.Option(help="Whether to use augmentations.")
    ] = True,
    independent_channels: Annotated[
        bool, typer.Option(help="Whether to train all channels independently.")
    ] = False,
    loss: Annotated[
        click.Choice,
        typer.Option(
            click_type=click.Choice(["mae", "mse"]),
            help="Loss function to use.",
        ),
    ] = "mae",
    n_channels_in: Annotated[
        Optional[int], typer.Option(help="Number of channels in")
    ] = None,
    n_channels_out: Annotated[
        Optional[int], typer.Option(help="Number of channels out")
    ] = None,
    logger: Annotated[
        click.Choice,
        typer.Option(
            click_type=click.Choice(["wandb", "tensorboard", "none"]),
            help="Logger to use.",
        ),
    ] = "none",
    # TODO: How to address model kwargs
) -> None:
    """
    Create a configuration for training Noise2Noise.

    If "Z" is present in `axes`, then `path_size` must be a list of length 3, otherwise
    2.

    If "C" is present in `axes`, then you need to set `n_channels` to the number of
    channels. Likewise, if you set the number of channels, then "C" must be present in
    `axes`.

    By default, all channels are trained together. To train all channels independently,
    set `independent_channels` to True.

    By setting `use_augmentations` to False, the only transformation applied will be
    normalization.
    """
    config = create_n2n_configuration(
        experiment_name=experiment_name,
        data_type=data_type,
        axes=axes,
        patch_size=patch_size,
        batch_size=batch_size,
        num_epochs=num_epochs,
        # TODO: fix choosing augmentations
        augmentations=None if use_augmentations else [],
        independent_channels=independent_channels,
        loss=loss,
        n_channels_in=n_channels_in,
        n_channels_out=n_channels_out,
        logger=logger,
    )
    _config_builder_exit(ctx, config)


@app.command()
def n2v(  # numpydoc ignore=PR01
    ctx: typer.Context,
    experiment_name: Annotated[str, typer.Option(help="Name of the experiment.")],
    axes: Annotated[str, typer.Option(help="Axes of the data (e.g. SYX).")],
    patch_size: Annotated[
        click.Tuple,
        typer.Option(
            help=(
                "Size of the patches along the spatial dimensions (if the data "
                "is not 3D pass the last value as -1 e.g. --patch-size 64 64 -1)."
            ),
            click_type=click.Tuple([int, int, int]),
            callback=handle_2D_3D_callback,
        ),
    ],
    batch_size: Annotated[int, typer.Option(help="Batch size.")],
    num_epochs: Annotated[int, typer.Option(help="Number of epochs.")],
    data_type: Annotated[
        click.Choice,
        typer.Option(click_type=click.Choice(["tiff"]), help="Type of the data."),
    ] = "tiff",
    use_augmentations: Annotated[
        bool, typer.Option(help="Whether to use augmentations.")
    ] = True,
    independent_channels: Annotated[
        bool, typer.Option(help="Whether to train all channels independently.")
    ] = True,
    use_n2v2: Annotated[bool, typer.Option(help="Whether to use N2V2")] = False,
    n_channels: Annotated[
        Optional[int], typer.Option(help="Number of channels (in and out)")
    ] = None,
    roi_size: Annotated[int, typer.Option(help="N2V pixel manipulation area.")] = 11,
    masked_pixel_percentage: Annotated[
        float, typer.Option(help="Percentage of pixels masked in each patch.")
    ] = 0.2,
    struct_n2v_axis: Annotated[
        click.Choice,
        typer.Option(click_type=click.Choice(["horizontal", "vertical", "none"])),
    ] = "none",
    struct_n2v_span: Annotated[
        int, typer.Option(help="Span of the structN2V mask.")
    ] = 5,
    logger: Annotated[
        click.Choice,
        typer.Option(
            click_type=click.Choice(["wandb", "tensorboard", "none"]),
            help="Logger to use.",
        ),
    ] = "none",
    # TODO: How to address model kwargs
) -> None:
    """
    Create a configuration for training Noise2Void.

    N2V uses a UNet model to denoise images in a self-supervised manner. To use its
    variants structN2V and N2V2, set the `struct_n2v_axis` and `struct_n2v_span`
    (structN2V) parameters, or set `use_n2v2` to True (N2V2).

    N2V2 modifies the UNet architecture by adding blur pool layers and removes the skip
    connections, thus removing checkboard artefacts. StructN2V is used when vertical
    or horizontal correlations are present in the noise; it applies an additional mask
    to the manipulated pixel neighbors.

    If "Z" is present in `axes`, then `path_size` must be a list of length 3, otherwise
    2.

    If "C" is present in `axes`, then you need to set `n_channels` to the number of
    channels.

    By default, all channels are trained independently. To train all channels together,
    set `independent_channels` to False.

    By setting `use_augmentations` to False, the only transformations applied will be
    normalization and N2V manipulation.

    The `roi_size` parameter specifies the size of the area around each pixel that will
    be manipulated by N2V. The `masked_pixel_percentage` parameter specifies how many
    pixels per patch will be manipulated.

    The parameters of the UNet can be specified in the `model_kwargs` (passed as a
    parameter-value dictionary). Note that `use_n2v2` and 'n_channels' override the
    corresponding parameters passed in `model_kwargs`.

    If you pass "horizontal" or "vertical" to `struct_n2v_axis`, then structN2V mask
    will be applied to each manipulated pixel.
    """
    config = create_n2v_configuration(
        experiment_name=experiment_name,
        data_type=data_type,
        axes=axes,
        patch_size=patch_size,
        batch_size=batch_size,
        num_epochs=num_epochs,
        # TODO: fix choosing augmentations
        augmentations=None if use_augmentations else [],
        independent_channels=independent_channels,
        use_n2v2=use_n2v2,
        n_channels=n_channels,
        roi_size=roi_size,
        masked_pixel_percentage=masked_pixel_percentage,
        struct_n2v_axis=struct_n2v_axis,
        struct_n2v_span=struct_n2v_span,
        logger=logger,
        # TODO: Model kwargs
    )
    _config_builder_exit(ctx, config)



"""
Convenience functions using torch.

These functions are used to control certain aspects and behaviours of PyTorch.
"""

import inspect
from typing import Union

import torch

from careamics.config.support import SupportedOptimizer, SupportedScheduler

from ..utils.logging import get_logger

logger = get_logger(__name__)  # TODO are logger still needed?


def filter_parameters(
    func: type,
    user_params: dict,
) -> dict:
    """
    Filter parameters according to the function signature.

    Parameters
    ----------
    func : type
        Class object.
    user_params : dict
        User provided parameters.

    Returns
    -------
    dict
        Parameters matching `func`'s signature.
    """
    # Get the list of all default parameters
    default_params = list(inspect.signature(func).parameters.keys())

    # Filter matching parameters
    params_to_be_used = set(user_params.keys()) & set(default_params)

    return {key: user_params[key] for key in params_to_be_used}


def get_optimizer(name: str) -> torch.optim.Optimizer:
    """
    Return the optimizer class given its name.

    Parameters
    ----------
    name : str
        Optimizer name.

    Returns
    -------
    torch.nn.Optimizer
        Optimizer class.
    """
    if name not in SupportedOptimizer:
        raise NotImplementedError(f"Optimizer {name} is not yet supported.")

    return getattr(torch.optim, name)


def get_optimizers() -> dict[str, str]:
    """
    Return the list of all optimizers available in torch.optim.

    Returns
    -------
    dict
        Optimizers available in torch.optim.
    """
    optims = {}
    for name, obj in inspect.getmembers(torch.optim):
        if inspect.isclass(obj) and issubclass(obj, torch.optim.Optimizer):
            if name != "Optimizer":
                optims[name] = name
    return optims


def get_scheduler(
    name: str,
) -> Union[
    # torch.optim.lr_scheduler.LRScheduler,
    torch.optim.lr_scheduler.ReduceLROnPlateau,
]:
    """
    Return the scheduler class given its name.

    Parameters
    ----------
    name : str
        Scheduler name.

    Returns
    -------
    Union
        Scheduler class.
    """
    if name not in SupportedScheduler:
        raise NotImplementedError(f"Scheduler {name} is not yet supported.")

    return getattr(torch.optim.lr_scheduler, name)


def get_schedulers() -> dict[str, str]:
    """
    Return the list of all schedulers available in torch.optim.lr_scheduler.

    Returns
    -------
    dict
        Schedulers available in torch.optim.lr_scheduler.
    """
    schedulers = {}
    for name, obj in inspect.getmembers(torch.optim.lr_scheduler):
        if inspect.isclass(obj) and issubclass(
            obj, torch.optim.lr_scheduler.LRScheduler
        ):
            if "LRScheduler" not in name:
                schedulers[name] = name
        elif name == "ReduceLROnPlateau":  # somewhat not a subclass of LRScheduler
            schedulers[name] = name
    return schedulers



import numpy as np
import pytest
import yaml

from careamics.config.data.data_model import DataConfig
from careamics.config.support import (
    SupportedTransform,
)
from careamics.config.transformations import NormalizeModel
from careamics.transforms import get_all_transforms


@pytest.mark.parametrize("ext", ["nd2", "jpg", "png ", "zarr", "npy"])
def test_wrong_extensions(minimum_data: dict, ext: str):
    """Test that supported model raises ValueError for unsupported extensions."""
    minimum_data["data_type"] = ext

    # instantiate DataModel model
    with pytest.raises(ValueError):
        DataConfig(**minimum_data)


@pytest.mark.parametrize("mean, std", [(0, 124.5), (12.6, 0.1)])
def test_mean_std_non_negative(minimum_data: dict, mean, std):
    """Test that non negative mean and std are accepted."""
    minimum_data["image_means"] = [mean]
    minimum_data["image_stds"] = [std]
    minimum_data["target_means"] = [mean]
    minimum_data["target_stds"] = [std]

    data_model = DataConfig(**minimum_data)
    assert data_model.image_means == [mean]
    assert data_model.image_stds == [std]
    assert data_model.target_means == [mean]
    assert data_model.target_stds == [std]


def test_mean_std_both_specified_or_none(minimum_data: dict):
    """Test an error is raised if std is specified but mean is None."""
    # No error if both are None
    DataConfig(**minimum_data)

    # Error if only mean is defined
    minimum_data["image_means"] = [10.4]
    with pytest.raises(ValueError):
        DataConfig(**minimum_data)

    # Error if only std is defined
    minimum_data.pop("image_means")
    minimum_data["image_stds"] = [10.4]
    with pytest.raises(ValueError):
        DataConfig(**minimum_data)

    # No error if both are specified
    minimum_data["image_means"] = [10.4]
    minimum_data["image_stds"] = [10.4]
    DataConfig(**minimum_data)

    # Error if target mean is defined but target std is None
    minimum_data["target_stds"] = [10.4, 11]
    with pytest.raises(ValueError):
        DataConfig(**minimum_data)


def test_set_mean_and_std(minimum_data: dict):
    """Test that mean and std can be set after initialization."""
    # they can be set both, when they None
    mean = [4.07]
    std = [14.07]
    data = DataConfig(**minimum_data)
    data.set_means_and_stds(mean, std)
    assert data.image_means == mean
    assert data.image_stds == std

    # Set also target mean and std
    data.set_means_and_stds(mean, std, mean, std)
    assert data.target_means == mean
    assert data.target_stds == std


def test_normalize_not_accepted(minimum_data: dict):
    """Test that normalize is not accepted, because it is mandatory and applied else
    where."""
    minimum_data["image_means"] = [10.4]
    minimum_data["image_stds"] = [3.2]
    minimum_data["transforms"] = [
        NormalizeModel(image_means=[0.485], image_stds=[0.229])
    ]

    with pytest.raises(ValueError):
        DataConfig(**minimum_data)


def test_patch_size(minimum_data: dict):
    """Test that non-zero even patch size are accepted."""
    # 2D
    data_model = DataConfig(**minimum_data)

    # 3D
    minimum_data["patch_size"] = [16, 8, 8]
    minimum_data["axes"] = "ZYX"

    data_model = DataConfig(**minimum_data)
    assert data_model.patch_size == minimum_data["patch_size"]


@pytest.mark.parametrize(
    "patch_size", [[12], [0, 12, 12], [12, 12, 13], [16, 10, 16], [12, 12, 12, 12]]
)
def test_wrong_patch_size(minimum_data: dict, patch_size):
    """Test that wrong patch sizes are not accepted (zero or odd, dims 1 or > 3)."""
    minimum_data["axes"] = "ZYX" if len(patch_size) == 3 else "YX"
    minimum_data["patch_size"] = patch_size

    with pytest.raises(ValueError):
        DataConfig(**minimum_data)


def test_set_3d(minimum_data: dict):
    """Test that 3D can be set."""
    data = DataConfig(**minimum_data)
    assert "Z" not in data.axes
    assert len(data.patch_size) == 2

    # error if changing Z manually
    with pytest.raises(ValueError):
        data.axes = "ZYX"

    # or patch size
    data = DataConfig(**minimum_data)
    with pytest.raises(ValueError):
        data.patch_size = [64, 64, 64]

    # set 3D
    data = DataConfig(**minimum_data)
    data.set_3D("ZYX", [64, 64, 64])
    assert "Z" in data.axes
    assert len(data.patch_size) == 3


def test_passing_empty_transforms(minimum_data: dict):
    """Test that empty list of transforms can be passed."""
    minimum_data["transforms"] = []
    DataConfig(**minimum_data)


def test_passing_incorrect_element(minimum_data: dict):
    """Test that incorrect element in the list of transforms raises an error (
    e.g. passing un object rather than a string)."""
    minimum_data["transforms"] = [
        {"name": get_all_transforms()[SupportedTransform.XY_FLIP.value]()},
    ]
    with pytest.raises(ValueError):
        DataConfig(**minimum_data)


def test_no_shuffle_in_train_dataloader_params(minimum_data: dict):
    """
    Test that an error is raised if there is no "shuffle" value in
    `train_dataloader_params`.
    """
    minimum_data["train_dataloader_params"] = {"num_workers": 4}
    with pytest.raises(ValueError):
        DataConfig(**minimum_data)


def test_export_to_yaml_float32_stats(tmp_path, minimum_data: dict):
    """Test exporting and loading the pydantic model when the statistics are
    np.float32."""
    data = DataConfig(**minimum_data)

    # set np.float32 stats values
    data.set_means_and_stds([np.float32(1234.5678)], [np.float32(21.73)])

    # export to yaml
    config_path = tmp_path / "data_config.yaml"
    with open(config_path, "w") as f:
        # dump configuration
        yaml.dump(data.model_dump(), f, default_flow_style=False, sort_keys=False)

    # load model
    dictionary = yaml.load(config_path.open("r"), Loader=yaml.SafeLoader)
    read_data = DataConfig(**dictionary)
    assert read_data.model_dump() == data.model_dump()



"""Module use to build BMZ model description."""

from pathlib import Path
from typing import Optional, Union

import numpy as np
from bioimageio.spec._internal.io import resolve_and_extract
from bioimageio.spec.model.v0_5 import (
    ArchitectureFromLibraryDescr,
    Author,
    AxisBase,
    AxisId,
    BatchAxis,
    ChannelAxis,
    EnvironmentFileDescr,
    FileDescr,
    FixedZeroMeanUnitVarianceAlongAxisKwargs,
    FixedZeroMeanUnitVarianceDescr,
    Identifier,
    InputTensorDescr,
    ModelDescr,
    OutputTensorDescr,
    PytorchStateDictWeightsDescr,
    SpaceInputAxis,
    SpaceOutputAxis,
    TensorId,
    Version,
    WeightsDescr,
)

from careamics.config import Configuration, DataConfig

from ._readme_factory import readme_factory


def _create_axes(
    array: np.ndarray,
    data_config: DataConfig,
    channel_names: Optional[list[str]] = None,
    is_input: bool = True,
) -> list[AxisBase]:
    """Create axes description.

    Array shape is expected to be SC(Z)YX.

    Parameters
    ----------
    array : np.ndarray
        Array.
    data_config : DataModel
        CAREamics data configuration.
    channel_names : Optional[list[str]], optional
        Channel names, by default None.
    is_input : bool, optional
        Whether the axes are input axes, by default True.

    Returns
    -------
    list[AxisBase]
        list of axes description.

    Raises
    ------
    ValueError
        If channel names are not provided when channel axis is present.
    """
    # axes have to be SC(Z)YX
    spatial_axes = data_config.axes.replace("S", "").replace("C", "")

    # batch is always present
    axes_model = [BatchAxis()]

    if "C" in data_config.axes:
        if channel_names is not None:
            axes_model.append(
                ChannelAxis(channel_names=[Identifier(name) for name in channel_names])
            )
        else:
            raise ValueError(
                f"Channel names must be provided if channel axis is present, axes: "
                f"{data_config.axes}."
            )
    else:
        # singleton channel
        axes_model.append(ChannelAxis(channel_names=[Identifier("channel")]))

    # spatial axes
    for ind, axes in enumerate(spatial_axes):
        if axes in ["X", "Y", "Z"]:
            if is_input:
                axes_model.append(
                    SpaceInputAxis(id=AxisId(axes.lower()), size=array.shape[2 + ind])
                )
            else:
                axes_model.append(
                    SpaceOutputAxis(id=AxisId(axes.lower()), size=array.shape[2 + ind])
                )

    return axes_model


def _create_inputs_ouputs(
    input_array: np.ndarray,
    output_array: np.ndarray,
    data_config: DataConfig,
    input_path: Union[Path, str],
    output_path: Union[Path, str],
    channel_names: Optional[list[str]] = None,
) -> tuple[InputTensorDescr, OutputTensorDescr]:
    """Create input and output tensor description.

    Input and output paths must point to a `.npy` file.

    Parameters
    ----------
    input_array : np.ndarray
        Input array.
    output_array : np.ndarray
        Output array.
    data_config : DataModel
        CAREamics data configuration.
    input_path : Union[Path, str]
        Path to input .npy file.
    output_path : Union[Path, str]
        Path to output .npy file.
    channel_names : Optional[list[str]], optional
        Channel names, by default None.

    Returns
    -------
    tuple[InputTensorDescr, OutputTensorDescr]
        Input and output tensor descriptions.
    """
    input_axes = _create_axes(input_array, data_config, channel_names)
    output_axes = _create_axes(output_array, data_config, channel_names, False)

    # mean and std
    assert data_config.image_means is not None, "Mean cannot be None."
    assert data_config.image_means is not None, "Std cannot be None."
    means = data_config.image_means
    stds = data_config.image_stds

    # and the mean and std required to invert the normalization
    # CAREamics denormalization: x = y * (std + eps) + mean
    # BMZ normalization : x = (y - mean') / (std' + eps)
    # to apply the BMZ normalization as a denormalization step, we need:
    eps = 1e-6
    inv_means = []
    inv_stds = []
    if means and stds:
        for mean, std in zip(means, stds):
            inv_means.append(-mean / (std + eps))
            inv_stds.append(1 / (std + eps) - eps)

        # create input/output descriptions
        input_descr = InputTensorDescr(
            id=TensorId("input"),
            axes=input_axes,
            test_tensor=FileDescr(source=input_path),
            preprocessing=[
                FixedZeroMeanUnitVarianceDescr(
                    kwargs=FixedZeroMeanUnitVarianceAlongAxisKwargs(
                        mean=means, std=stds, axis="channel"
                    )
                )
            ],
        )
        output_descr = OutputTensorDescr(
            id=TensorId("prediction"),
            axes=output_axes,
            test_tensor=FileDescr(source=output_path),
            postprocessing=[
                FixedZeroMeanUnitVarianceDescr(
                    kwargs=FixedZeroMeanUnitVarianceAlongAxisKwargs(  # invert norm
                        mean=inv_means, std=inv_stds, axis="channel"
                    )
                )
            ],
        )

        return input_descr, output_descr
    else:
        raise ValueError("Mean and std cannot be None.")


def create_model_description(
    config: Configuration,
    name: str,
    general_description: str,
    data_description: str,
    authors: list[Author],
    inputs: Union[Path, str],
    outputs: Union[Path, str],
    weights_path: Union[Path, str],
    torch_version: str,
    careamics_version: str,
    config_path: Union[Path, str],
    env_path: Union[Path, str],
    covers: list[Union[Path, str]],
    channel_names: Optional[list[str]] = None,
    model_version: str = "0.1.0",
) -> ModelDescr:
    """Create model description.

    Parameters
    ----------
    config : Configuration
        CAREamics configuration.
    name : str
        Name of the model.
    general_description : str
        General description of the model.
    data_description : str
        Description of the data the model was trained on.
    authors : list[Author]
        Authors of the model.
    inputs : Union[Path, str]
        Path to input .npy file.
    outputs : Union[Path, str]
        Path to output .npy file.
    weights_path : Union[Path, str]
        Path to model weights.
    torch_version : str
        Pytorch version.
    careamics_version : str
        CAREamics version.
    config_path : Union[Path, str]
        Path to model configuration.
    env_path : Union[Path, str]
        Path to environment file.
    covers : list of pathlib.Path or str
        Paths to cover images.
    channel_names : Optional[list[str]], optional
        Channel names, by default None.
    model_version : str, default "0.1.0"
        Model version.

    Returns
    -------
    ModelDescr
        Model description.
    """
    # documentation
    doc = readme_factory(
        config,
        careamics_version=careamics_version,
        data_description=data_description,
    )

    # inputs, outputs
    input_descr, output_descr = _create_inputs_ouputs(
        input_array=np.load(inputs),
        output_array=np.load(outputs),
        data_config=config.data_config,
        input_path=inputs,
        output_path=outputs,
        channel_names=channel_names,
    )

    # weights description
    architecture_descr = ArchitectureFromLibraryDescr(
        import_from="careamics.models.unet",
        callable=f"{config.algorithm_config.model.architecture}",
        kwargs=config.algorithm_config.model.model_dump(),
    )

    weights_descr = WeightsDescr(
        pytorch_state_dict=PytorchStateDictWeightsDescr(
            source=weights_path,
            architecture=architecture_descr,
            pytorch_version=Version(torch_version),
            dependencies=EnvironmentFileDescr(source=env_path),
        ),
    )

    # overall model description
    model = ModelDescr(
        name=name,
        authors=authors,
        description=general_description,
        documentation=doc,
        inputs=[input_descr],
        outputs=[output_descr],
        tags=config.get_algorithm_keywords(),
        links=[
            "https://github.com/CAREamics/careamics",
            "https://careamics.github.io/latest/",
        ],
        license="BSD-3-Clause",
        config={
            "bioimageio": {
                "test_kwargs": {
                    "pytorch_state_dict": {
                        "absolute_tolerance": 1e-2,
                        "relative_tolerance": 1e-2,
                    }
                }
            }
        },
        version=model_version,
        weights=weights_descr,
        attachments=[FileDescr(source=config_path)],
        cite=config.get_algorithm_citations(),
        covers=covers,
    )

    return model


def extract_model_path(model_desc: ModelDescr) -> tuple[Path, Path]:
    """Return the relative path to the weights and configuration files.

    Parameters
    ----------
    model_desc : ModelDescr
        Model description.

    Returns
    -------
    tuple of (path, path)
        Weights and configuration paths.
    """
    if model_desc.weights.pytorch_state_dict is None:
        raise ValueError("No model weights found in model description.")
    weights_path = resolve_and_extract(
        model_desc.weights.pytorch_state_dict.source
    ).path

    for file in model_desc.attachments:
        file_path = file.source if isinstance(file.source, Path) else file.source.path
        if file_path is None:
            continue
        file_path = Path(file_path)
        if file_path.name == "careamics.yaml":
            config_path = resolve_and_extract(file.source).path
            break
    else:
        raise ValueError("Configuration file not found.")

    return weights_path, config_path



"""Dataset module."""

__all__ = [
    "InMemoryDataset",
    "InMemoryPredDataset",
    "InMemoryTiledPredDataset",
    "IterablePredDataset",
    "IterableTiledPredDataset",
    "PathIterableDataset",
]

from .in_memory_dataset import InMemoryDataset
from .in_memory_pred_dataset import InMemoryPredDataset
from .in_memory_tiled_pred_dataset import InMemoryTiledPredDataset
from .iterable_dataset import PathIterableDataset
from .iterable_pred_dataset import IterablePredDataset
from .iterable_tiled_pred_dataset import IterableTiledPredDataset



"""Prediction utility functions."""

import builtins
from typing import Union

import numpy as np
from numpy.typing import NDArray

from careamics.config.tile_information import TileInformation


# TODO: why not allow input and output of torch.tensor ?
def stitch_prediction(
    tiles: list[np.ndarray],
    tile_infos: list[TileInformation],
) -> list[np.ndarray]:
    """
    Stitch tiles back together to form a full image(s).

    Tiles are of dimensions SC(Z)YX, where C is the number of channels and can be a
    singleton dimension.

    Parameters
    ----------
    tiles : list of numpy.ndarray
        Cropped tiles and their respective stitching coordinates. Can contain tiles
        from multiple images.
    tile_infos : list of TileInformation
        List of information and coordinates obtained from
        `dataset.tiled_patching.extract_tiles`.

    Returns
    -------
    list of numpy.ndarray
        Full image(s).
    """
    # Find where to split the lists so that only info from one image is contained.
    # Do this by locating the last tiles of each image.
    last_tiles = [tile_info.last_tile for tile_info in tile_infos]
    last_tile_position = np.where(last_tiles)[0]
    image_slices = [
        slice(
            None if i == 0 else last_tile_position[i - 1] + 1, last_tile_position[i] + 1
        )
        for i in range(len(last_tile_position))
    ]
    image_predictions = []
    # slice the lists and apply stitch_prediction_single to each in turn.
    for image_slice in image_slices:
        image_predictions.append(
            stitch_prediction_single(tiles[image_slice], tile_infos[image_slice])
        )
    return image_predictions


def stitch_prediction_single(
    tiles: list[NDArray],
    tile_infos: list[TileInformation],
) -> NDArray:
    """
    Stitch tiles back together to form a full image.

    Tiles are of dimensions SC(Z)YX, where C is the number of channels and can be a
    singleton dimension.

    Parameters
    ----------
    tiles : list of numpy.ndarray
        Cropped tiles and their respective stitching coordinates.
    tile_infos : list of TileInformation
        List of information and coordinates obtained from
        `dataset.tiled_patching.extract_tiles`.

    Returns
    -------
    numpy.ndarray
        Full image, with dimensions SC(Z)YX.
    """
    # TODO: this is hacky... need a better way to deal with when input channels and
    #   target channels do not match
    if len(tile_infos[0].array_shape) == 4:
        # 4 dimensions => 3 spatial dimensions so -4 is channel dimension
        tile_channels = tiles[0].shape[-4]
    elif len(tile_infos[0].array_shape) == 3:
        # 3 dimensions => 2 spatial dimensions so -3 is channel dimension
        tile_channels = tiles[0].shape[-3]
    else:
        # Note pretty sure this is unreachable because array shape is already
        #   validated by TileInformation
        raise ValueError(
            f"Unsupported number of output dimension {len(tile_infos[0].array_shape)}"
        )
    # retrieve whole array size, add S dim and use number of channels in tile
    input_shape = (1, tile_channels, *tile_infos[0].array_shape[1:])
    predicted_image = np.zeros(input_shape, dtype=np.float32)

    for tile, tile_info in zip(tiles, tile_infos):

        # Compute coordinates for cropping predicted tile
        crop_slices: tuple[Union[builtins.ellipsis, slice], ...] = (
            ...,
            *[slice(c[0], c[1]) for c in tile_info.overlap_crop_coords],
        )

        # Crop predited tile according to overlap coordinates
        cropped_tile = tile[crop_slices]

        # Insert cropped tile into predicted image using stitch coordinates
        image_slices = (..., *[slice(c[0], c[1]) for c in tile_info.stitch_coords])
        predicted_image[image_slices] = cropped_tile.astype(np.float32)

    return predicted_image



import pytest

from careamics.config.transformations import N2VManipulateModel
from careamics.transforms import N2VManipulate


def test_odd_roi_and_mask():
    """Test that errors are thrown if we pass even roi and mask sizes."""
    # no error
    model = N2VManipulateModel(roi_size=3, struct_mask_span=7)
    assert model.roi_size == 3
    assert model.struct_mask_span == 7

    # errors
    with pytest.raises(ValueError):
        N2VManipulateModel(roi_size=4, struct_mask_span=7)

    with pytest.raises(ValueError):
        N2VManipulateModel(roi_size=3, struct_mask_span=6)


def test_extra_parameters():
    """Test that errors are thrown if we pass extra parameters."""
    with pytest.raises(ValueError):
        N2VManipulateModel(roi_size=3, struct_mask_span=7, extra_param=1)


def test_comptatibility_with_transform():
    """Test that the model allows instantiating a transform."""
    model = N2VManipulateModel(roi_size=3, struct_mask_span=7, strategy="median")

    # instantiate transform
    N2VManipulate(**model.model_dump())



"""Supported configuration options.

Used throughout the code to ensure consistency. These should be kept in sync with the
corresponding configuration options in the Pydantic models.
"""

__all__ = [
    "SupportedActivation",
    "SupportedAlgorithm",
    "SupportedArchitecture",
    "SupportedData",
    "SupportedLogger",
    "SupportedLoss",
    "SupportedOptimizer",
    "SupportedPixelManipulation",
    "SupportedScheduler",
    "SupportedStructAxis",
    "SupportedTransform",
]


from .supported_activations import SupportedActivation
from .supported_algorithms import SupportedAlgorithm
from .supported_architectures import SupportedArchitecture
from .supported_data import SupportedData
from .supported_loggers import SupportedLogger
from .supported_losses import SupportedLoss
from .supported_optimizers import SupportedOptimizer, SupportedScheduler
from .supported_pixel_manipulations import SupportedPixelManipulation
from .supported_struct_axis import SupportedStructAxis
from .supported_transforms import SupportedTransform



import pytest

from careamics.config.transformations import NormalizeModel
from careamics.transforms import Normalize


def test_setting_image_means_std():
    """Test that we can set the image_means and std values."""
    model = NormalizeModel(image_means=[0.5], image_stds=[0.5])
    assert model.image_means == [0.5]
    assert model.image_stds == [0.5]

    model.image_means = [0.6]
    model.image_stds = [0.6]
    assert model.image_means == [0.6]
    assert model.image_stds == [0.6]

    model = NormalizeModel(
        image_means=[0.5],
        image_stds=[0.5],
        target_means=[0.5],
        target_stds=[0.5],
    )
    assert model.image_means == [0.5]
    assert model.image_stds == [0.5]
    assert model.target_means == [0.5]
    assert model.target_stds == [0.5]

    model.image_means = [0.6]
    model.image_stds = [0.6]
    model.target_means = [0.6]
    model.target_stds = [0.6]

    assert model.image_means == [0.6]
    assert model.image_stds == [0.6]
    assert model.target_means == [0.6]
    assert model.target_stds == [0.6]


def test_error_different_length_means_stds():
    """Test that an error is raised if the image_means and stds have different
    lengths."""
    with pytest.raises(ValueError):
        NormalizeModel(image_means=[0.5], image_stds=[0.5, 0.6])

    with pytest.raises(ValueError):
        NormalizeModel(image_means=[0.5, 0.6], image_stds=[0.5])

    with pytest.raises(ValueError):
        NormalizeModel(
            image_means=[0.5],
            image_stds=[0.5],
            target_means=[0.5],
        )

    with pytest.raises(ValueError):
        NormalizeModel(
            image_means=[0.5],
            image_stds=[0.5],
            target_stds=[0.5],
        )

    with pytest.raises(ValueError):
        NormalizeModel(
            image_means=[0.5],
            image_stds=[0.5],
            target_means=[0.5, 0.6],
            target_stds=[0.5],
        )

    with pytest.raises(ValueError):
        NormalizeModel(
            image_means=[0.5],
            image_stds=[0.5],
            target_means=[0.5],
            target_stds=[0.5, 0.6],
        )


def test_comptatibility_with_transform():
    """Test that the model allows instantiating a transform."""
    model = NormalizeModel(image_means=[0.5], image_stds=[0.5])

    # instantiate transform
    Normalize(**model.model_dump())



"""Patch transform applying XY random 90 degrees rotations."""

from typing import Optional

import numpy as np
from numpy.typing import NDArray

from careamics.transforms.transform import Transform


class XYRandomRotate90(Transform):
    """Applies random 90 degree rotations to the YX axis.

    This transform expects C(Z)YX dimensions.

    Attributes
    ----------
    rng : np.random.Generator
        Random number generator.
    p : float
        Probability of applying the transform.
    seed : Optional[int]
        Random seed.

    Parameters
    ----------
    p : float
        Probability of applying the transform, by default 0.5.
    seed : Optional[int]
        Random seed, by default None.
    """

    def __init__(self, p: float = 0.5, seed: Optional[int] = None):
        """Constructor.

        Parameters
        ----------
        p : float
            Probability of applying the transform, by default 0.5.
        seed : Optional[int]
            Random seed, by default None.
        """
        if p < 0 or p > 1:
            raise ValueError("Probability must be in [0, 1].")

        # probability to apply the transform
        self.p = p

        # numpy random generator
        self.rng = np.random.default_rng(seed=seed)

    def __call__(
        self,
        patch: NDArray,
        target: Optional[NDArray] = None,
        **additional_arrays: NDArray,
    ) -> tuple[NDArray, Optional[NDArray], dict[str, NDArray]]:
        """Apply the transform to the source patch and the target (optional).

        Parameters
        ----------
        patch : np.ndarray
            Patch, 2D or 3D, shape C(Z)YX.
        target : Optional[np.ndarray], optional
            Target for the patch, by default None.
        **additional_arrays : NDArray
            Additional arrays that will be transformed identically to `patch` and
            `target`.

        Returns
        -------
        tuple[np.ndarray, Optional[np.ndarray]]
            Transformed patch and target.
        """
        if self.rng.random() > self.p:
            return patch, target, additional_arrays

        # number of rotations
        n_rot = self.rng.integers(1, 4)

        axes = (-2, -1)
        patch_transformed = self._apply(patch, n_rot, axes)
        target_transformed = (
            self._apply(target, n_rot, axes) if target is not None else None
        )
        additional_transformed = {
            key: self._apply(array, n_rot, axes)
            for key, array in additional_arrays.items()
        }

        return patch_transformed, target_transformed, additional_transformed

    def _apply(self, patch: NDArray, n_rot: int, axes: tuple[int, int]) -> NDArray:
        """Apply the transform to the image.

        Parameters
        ----------
        patch : np.ndarray
            Image or image patch, 2D or 3D, shape C(Z)YX.
        n_rot : int
            Number of 90 degree rotations.
        axes : tuple[int, int]
            Axes along which to rotate the patch.

        Returns
        -------
        np.ndarray
            Transformed patch.
        """
        return np.ascontiguousarray(np.rot90(patch, k=n_rot, axes=axes))



"""Architecture model validators."""

from careamics.config.architectures import UNetModel


def model_without_n2v2(model: UNetModel) -> UNetModel:
    """Validate that the Unet model does not have the n2v2 attribute.

    Parameters
    ----------
    model : UNetModel
        Model to validate.

    Returns
    -------
    UNetModel
        The validated model.

    Raises
    ------
    ValueError
        If the model has the `n2v2` attribute set to `True`.
    """
    if model.n2v2:
        raise ValueError(
            "The algorithm does not support the `n2v2` attribute in the model. "
            "Set it to `False`."
        )

    return model


def model_without_final_activation(model: UNetModel) -> UNetModel:
    """Validate that the UNet model does not have the final_activation.

    Parameters
    ----------
    model : UNetModel
        Model to validate.

    Returns
    -------
    UNetModel
        The validated model.

    Raises
    ------
    ValueError
        If the model has the final_activation attribute set.
    """
    if model.final_activation != "None":
        raise ValueError(
            "The algorithm does not support a `final_activation` in the model. "
            'Set it to `"None"`.'
        )

    return model


def model_matching_in_out_channels(model: UNetModel) -> UNetModel:
    """Validate that the UNet model has the same number of channel inputs and outputs.

    Parameters
    ----------
    model : UNetModel
        Model to validate.

    Returns
    -------
    UNetModel
        Validated model.

    Raises
    ------
    ValueError
        If the model has different number of input and output channels.
    """
    if model.num_classes != model.in_channels:
        raise ValueError(
            "The algorithm requires the same number of input and output channels. "
            "Make sure that `in_channels` and `num_classes` are equal."
        )

    return model



# %%
from collections.abc import Sequence
from pathlib import Path
from typing import TypedDict

import matplotlib.pyplot as plt
import numpy as np
import zarr
from numpy.typing import NDArray
from zarr.storage import FSStore

from careamics.config import DataConfig
from careamics.config.support import SupportedData
from careamics.dataset_ng.patch_extractor import create_patch_extractor
from careamics.dataset_ng.patch_extractor.image_stack import ZarrImageStack
from careamics.dataset_ng.patch_extractor.image_stack_loader import ImageStackLoader


# %%
def create_zarr_array(file_path: Path, data_path: str, data: NDArray):
    store = FSStore(url=file_path.resolve())
    # create array
    array = zarr.create(
        store=store,
        shape=data.shape,
        chunks=data.shape,  # only 1 chunk
        dtype=np.uint16,
        path=data_path,
    )
    # write data
    array[...] = data
    store.close()


def create_zarr(file_path: Path, data_paths: Sequence[str], data: Sequence[NDArray]):
    for data_path, array in zip(data_paths, data):
        create_zarr_array(file_path=file_path, data_path=data_path, data=array)


# %% [markdown]
# ### Create example ZARR file

# %%
dir_path = Path("/home/melisande.croft/Documents/Data")
file_name = "test_ngff_image.zarr"
file_path = dir_path / file_name

data_paths = [
    "image_1",
    "group_1/image_1.1",
    "group_1/image_1.2",
]
data_shapes = [(1, 3, 64, 64), (1, 3, 32, 48), (1, 3, 32, 32)]
data = [np.random.randint(1, 255, size=shape, dtype=np.uint8) for shape in data_shapes]
if not file_path.is_file() and not file_path.is_dir():
    create_zarr(file_path, data_paths, data)

# %% [markdown]
# ### Make sure file exists

# %%
store = FSStore(url=file_path.resolve(), mode="r")

# %%
list(store.keys())

# %% [markdown]
# ### Define custom loading function


# %%
class ZarrSource(TypedDict):
    store: FSStore
    data_paths: Sequence[str]


def custom_image_stack_loader(source: ZarrSource, axes: str, *args, **kwargs):
    image_stacks = [
        ZarrImageStack(store=source["store"], data_path=data_path, axes=axes)
        for data_path in source["data_paths"]
    ]
    return image_stacks


# %% [markdown]
# ### Test custom loading func

# %%
# dummy data config
data_config = DataConfig(data_type="custom", patch_size=[64, 64], axes="SCYX")

# %%
image_stack_loader: ImageStackLoader = custom_image_stack_loader

# %%
# So pylance knows that datatype is custom to match function overloads
assert data_config.data_type is SupportedData.CUSTOM

patch_extractor = create_patch_extractor(
    source={"store": store, "data_paths": data_paths},
    axes=data_config.axes,
    data_type=data_config.data_type,
    image_stack_loader=custom_image_stack_loader,
)

# %%
# extract patch and display
patch = patch_extractor.extract_patch(2, 0, (8, 16), (16, 16))
plt.imshow(np.moveaxis(patch, 0, -1))

# %%
